@@ -75,7 +75,7 @@ static void loadColumn(WMBrowser * bPtr, int column);
 
 static void removeColumn(WMBrowser * bPtr, int column);
 
-static char *createTruncatedString(WMFont * font, char *text, int *textLen, int width);
+static char *createTruncatedString(WMFont * font, const char *text, int *textLen, int width);
 
 static void willResizeBrowser(W_ViewDelegate *, WMView *, unsigned int *, unsigned int *);
 
@@ -220,7 +220,7 @@ int WMGetBrowserNumberOfColumns(WMBrowser * bPtr)
 	return bPtr->usedColumnCount;
 }
 
-void WMSetBrowserPathSeparator(WMBrowser * bPtr, char *separator)
+void WMSetBrowserPathSeparator(WMBrowser * bPtr, const char *separator)
 {
 	if (bPtr->pathSeparator)
 		wfree(bPtr->pathSeparator);
@@ -350,7 +350,7 @@ int WMGetBrowserSelectedRowInColumn(WMBrowser * bPtr, int column)
 	}
 }
 
-void WMSetBrowserColumnTitle(WMBrowser * bPtr, int column, char *title)
+void WMSetBrowserColumnTitle(WMBrowser * bPtr, int column, const char *title)
 {
 	assert(column >= 0);
 	assert(column < bPtr->usedColumnCount);
@@ -375,8 +375,6 @@ void WMSetBrowserTitled(WMBrowser * bPtr, Bool flag)
 	if (bPtr->flags.isTitled == flag)
 		return;
 
-	columnX = 0;
-
 	if (!bPtr->flags.isTitled) {
 		columnY = TITLE_SPACING + bPtr->titleHeight;
 
@@ -414,7 +412,7 @@ void WMSortBrowserColumnWithComparer(WMBrowser * bPtr, int column, WMCompareData
 	WMSortListItemsWithComparer(bPtr->columns[column], func);
 }
 
-WMListItem *WMInsertBrowserItem(WMBrowser * bPtr, int column, int row, char *text, Bool isBranch)
+WMListItem *WMInsertBrowserItem(WMBrowser * bPtr, int column, int row, const char *text, Bool isBranch)
 {
 	WMListItem *item;
 
@@ -434,6 +432,9 @@ static void willResizeBrowser(W_ViewDelegate * self, WMView * view, unsigned int
 	int colX, colY;
 	int i;
 
+	/* Parameter not used, but tell the compiler that it is ok */
+	(void) self;
+
 	assert(*width > 0);
 	assert(*height > 0);
 
@@ -475,19 +476,24 @@ static void paintItem(WMList * lPtr, int index, Drawable d, char *text, int stat
 	Display *display = scr->display;
 	WMFont *font = ((state & WLDSIsBranch) ? scr->boldFont : scr->normalFont);
 	WMColor *backColor = ((state & WLDSSelected) ? scr->white : view->backColor);
-	int width, height, x, y, textLen;
+	int width, height, x, y;
+
+	/* Parameter not used, but tell the compiler that it is ok */
+	(void) index;
 
 	width = rect->size.width;
 	height = rect->size.height;
 	x = rect->pos.x;
 	y = rect->pos.y;
-	textLen = strlen(text);
 
 	XFillRectangle(display, d, WMColorGC(backColor), x, y, width, height);
 
 	if (text) {
+		int widthC, textLen;
+
 		/* Avoid overlaping... */
-		int widthC = (state & WLDSIsBranch) ? width - 20 : width - 8;
+		widthC = (state & WLDSIsBranch) ? width - 20 : width - 8;
+		textLen = strlen(text);
 		if (WMWidthOfString(font, text, textLen) > widthC) {
 			char *textBuf = createTruncatedString(font, text, &textLen, widthC);
 			W_PaintText(view, d, font, x + 4, y, widthC, WALeft, scr->black, False, textBuf, textLen);
@@ -559,7 +565,7 @@ static void scrollCallback(WMWidget * scroller, void *self)
 
 			floatValue = (floatValue * value) / value;
 
-			newFirst = rint(floatValue * (float)(bPtr->columnCount - bPtr->maxVisibleColumns));
+			newFirst = rint(floatValue * (double)(bPtr->columnCount - bPtr->maxVisibleColumns));
 
 			if (bPtr->firstVisibleColumn != newFirst)
 				scrollToColumn(bPtr, newFirst, False);
@@ -775,10 +781,7 @@ WMArray *WMGetBrowserPaths(WMBrowser * bPtr)
 		path = wmalloc(slen);
 		/* ignore first `/' */
 		for (i = 0; i <= column; i++) {
-			if (wstrlcat(path, bPtr->pathSeparator, slen) >= slen) {
-				wfree(path);
-				return NULL;
-			}
+			wstrlcat(path, bPtr->pathSeparator, slen);
 			if (i == column) {
 				item = lastItem;
 			} else {
@@ -786,10 +789,7 @@ WMArray *WMGetBrowserPaths(WMBrowser * bPtr)
 			}
 			if (!item)
 				break;
-			if (wstrlcat(path, item->text, slen) >= slen) {
-				wfree(path);
-				return NULL;
-			}
+			wstrlcat(path, item->text, slen);
 		}
 		WMAddToArray(paths, path);
 	}
@@ -1116,7 +1116,7 @@ static void destroyBrowser(WMBrowser * bPtr)
 	wfree(bPtr);
 }
 
-static char *createTruncatedString(WMFont * font, char *text, int *textLen, int width)
+static char *createTruncatedString(WMFont * font, const char *text, int *textLen, int width)
 {
 	size_t slen;
 	int dLen;
