@@ -20,6 +20,7 @@
 
 #include "wconfig.h"
 
+#include <stdint.h>
 #include <stdlib.h>
 #include <string.h>
 #include <sys/time.h>
@@ -27,18 +28,16 @@
 #include "WindowMaker.h"
 #include "screen.h"
 #include "framewin.h"
+#include "icon.h"
 #include "window.h"
 #include "defaults.h"
 #include "switchpanel.h"
-#include "funcs.h"
+#include "misc.h"
 #include "xinerama.h"
 
-extern Atom _XA_WM_IGNORE_FOCUS_EVENTS;
 
-#ifdef SHAPE
+#ifdef USE_XSHAPE
 #include <X11/extensions/shape.h>
-
-extern Bool wShapeSupported;
 #endif
 
 struct SwitchPanel {
@@ -49,6 +48,7 @@ struct SwitchPanel {
 	WMArray *icons;
 	WMArray *images;
 	WMArray *windows;
+	WMArray *flags;
 	RImage *bg;
 	int current;
 	int firstVisible;
@@ -56,8 +56,6 @@ struct SwitchPanel {
 
 	WMLabel *label;
 
-	RImage *defIcon;
-
 	RImage *tileTmp;
 	RImage *tile;
 
@@ -65,16 +63,16 @@ struct SwitchPanel {
 	WMColor *white;
 };
 
-extern WPreferences wPreferences;
-
 #define BORDER_SPACE 10
 #define ICON_SIZE 48
 #define ICON_TILE_SIZE 64
 #define LABEL_HEIGHT 25
 #define SCREEN_BORDER_SPACING 2*20
-#define SCROLL_STEPS (ICON_TILE_SIZE/2)
 
-static int canReceiveFocus(WWindow * wwin)
+#define ICON_SELECTED (1<<1)
+#define ICON_DIM (1<<2)
+
+static int canReceiveFocus(WWindow *wwin)
 {
 	if (wwin->frame->workspace != wwin->screen_ptr->current_workspace)
 		return 0;
@@ -96,19 +94,47 @@ static int canReceiveFocus(WWindow * wwin)
 	return 1;
 }
 
-static void changeImage(WSwitchPanel * panel, int idecks, int selected)
+static Bool sameWindowClass(WWindow *wwin, WWindow *curwin)
 {
-	WMFrame *icon = WMGetFromArray(panel->icons, idecks);
-	RImage *image = WMGetFromArray(panel->images, idecks);
+	if (!wwin->wm_class || !curwin->wm_class)
+		return False;
+	if (strcmp(wwin->wm_class, curwin->wm_class))
+		return False;
 
-	if (!panel->bg && !panel->tile) {
-		if (!selected)
-			WMSetFrameRelief(icon, WRFlat);
-	}
+	return True;
+}
+
+static void changeImage(WSwitchPanel *panel, int idecks, int selected, Bool dim, Bool force)
+{
+	WMFrame *icon = NULL;
+	RImage *image = NULL;
+	int flags;
+	int desired = 0;
+
+	/* This whole function is a no-op if we aren't drawing the panel */
+	if (!wPreferences.swtileImage)
+		return;
+
+	icon = WMGetFromArray(panel->icons, idecks);
+	image = WMGetFromArray(panel->images, idecks);
+	flags = (int) (uintptr_t) WMGetFromArray(panel->flags, idecks);
+
+	if (selected)
+		desired |= ICON_SELECTED;
+	if (dim)
+		desired |= ICON_DIM;
+
+	if (flags == desired && !force)
+		return;
+
+	WMReplaceInArray(panel->flags, idecks, (void *) (uintptr_t) desired);
+
+	if (!panel->bg && !panel->tile && !selected)
+		WMSetFrameRelief(icon, WRFlat);
 
 	if (image && icon) {
 		RImage *back;
-		int opaq = 255;
+		int opaq = (dim) ? 75 : 255;
 		RImage *tile;
 		WMPoint pos;
 		Pixmap p;
@@ -118,11 +144,11 @@ static void changeImage(WSwitchPanel * panel, int idecks, int selected)
 
 		pos = WMGetViewPosition(WMWidgetView(icon));
 		back = panel->tileTmp;
-		if (panel->bg)
+		if (panel->bg) {
 			RCopyArea(back, panel->bg,
 				  BORDER_SPACE + pos.x - panel->firstVisible * ICON_TILE_SIZE,
 				  BORDER_SPACE + pos.y, back->width, back->height, 0, 0);
-		else {
+		} else {
 			RColor color;
 			WMScreen *wscr = WMWidgetScreen(icon);
 			color.red = 255;
@@ -131,11 +157,13 @@ static void changeImage(WSwitchPanel * panel, int idecks, int selected)
 			color.blue = WMBlueComponentOfColor(WMGrayColor(wscr)) >> 8;
 			RFillImage(back, &color);
 		}
+
 		if (selected) {
 			tile = panel->tile;
 			RCombineArea(back, tile, 0, 0, tile->width, tile->height,
 				     (back->width - tile->width) / 2, (back->height - tile->height) / 2);
 		}
+
 		RCombineAreaWithOpaqueness(back, image, 0, 0, image->width, image->height,
 					   (back->width - image->width) / 2, (back->height - image->height) / 2,
 					   opaq);
@@ -146,24 +174,11 @@ static void changeImage(WSwitchPanel * panel, int idecks, int selected)
 		XFreePixmap(dpy, p);
 	}
 
-	if (!panel->bg && !panel->tile) {
-		if (selected)
-			WMSetFrameRelief(icon, WRSimple);
-	}
+	if (!panel->bg && !panel->tile && selected)
+		WMSetFrameRelief(icon, WRSimple);
 }
 
-static RImage *scaleDownIfNeeded(RImage * image)
-{
-	if (image && ((image->width - ICON_SIZE) > 2 || (image->height - ICON_SIZE) > 2)) {
-		RImage *nimage;
-		nimage = RScaleImage(image, ICON_SIZE, (image->height * ICON_SIZE / image->width));
-		RReleaseImage(image);
-		image = nimage;
-	}
-	return image;
-}
-
-static void addIconForWindow(WSwitchPanel * panel, WMWidget * parent, WWindow * wwin, int x, int y)
+static void addIconForWindow(WSwitchPanel *panel, WMWidget *parent, WWindow *wwin, int x, int y)
 {
 	WMFrame *icon = WMCreateFrame(parent);
 	RImage *image = NULL;
@@ -175,41 +190,23 @@ static void addIconForWindow(WSwitchPanel * panel, WMWidget * parent, WWindow *
 	if (!WFLAGP(wwin, always_user_icon) && wwin->net_icon_image)
 		image = RRetainImage(wwin->net_icon_image);
 
-	// Make this use a caching thing. When there are many windows,
-	// it's very likely that most of them are instances of the same thing,
-	// so caching them should get performance acceptable in these cases.
+	/* get_icon_image() includes the default icon image */
 	if (!image)
-		image = wDefaultGetImage(panel->scr, wwin->wm_instance, wwin->wm_class, ICON_TILE_SIZE);
+		image = get_icon_image(panel->scr, wwin->wm_instance, wwin->wm_class, ICON_TILE_SIZE);
 
-	if (!image && !panel->defIcon) {
-		char *file = wDefaultGetIconFile(panel->scr, NULL, NULL, False);
-		if (file) {
-			char *path = FindImage(wPreferences.icon_path, file);
-			if (path) {
-				image = RLoadImage(panel->scr->rcontext, path, 0);
-				wfree(path);
-			}
-		}
-		if (image)
-			panel->defIcon = scaleDownIfNeeded(image);
-		image = NULL;
-	}
-	if (!image && panel->defIcon)
-		image = RRetainImage(panel->defIcon);
-
-	image = scaleDownIfNeeded(image);
+	/* We must resize the icon size (~64) to the switch panel icon size (~48) */
+	image = wIconValidateIconSize(image, ICON_SIZE);
 
 	WMAddToArray(panel->images, image);
 	WMAddToArray(panel->icons, icon);
 }
 
-static void scrollIcons(WSwitchPanel * panel, int delta)
+static void scrollIcons(WSwitchPanel *panel, int delta)
 {
 	int nfirst = panel->firstVisible + delta;
 	int i;
 	int count = WMGetArrayItemCount(panel->windows);
-//    int nx, ox;
-//    struct timeval tv1, tv2;
+	Bool dim;
 
 	if (count <= panel->visibleCount)
 		return;
@@ -221,26 +218,16 @@ static void scrollIcons(WSwitchPanel * panel, int delta)
 
 	if (nfirst == panel->firstVisible)
 		return;
-/*
-    ox = -panel->firstVisible * ICON_TILE_SIZE;
-    nx = -nfirst * ICON_TILE_SIZE;
-    for (i= 0; i < SCROLL_STEPS; i++) {
-        unsigned int diff;
-        gettimeofday(&tv1, NULL);
-        WMMoveWidget(panel->iconBox, (nx*i + ox*(SCROLL_STEPS-i))/(SCROLL_STEPS-1), 0);
-        XSync(dpy, False);
-        gettimeofday(&tv2, NULL);
-        diff = (tv2.tv_sec-tv1.tv_sec)*10000+(tv2.tv_usec-tv1.tv_usec)/100;
-        if (diff < 200)
-          wusleep(300-diff);
-    }
- */
+
 	WMMoveWidget(panel->iconBox, -nfirst * ICON_TILE_SIZE, 0);
 
 	panel->firstVisible = nfirst;
 
 	for (i = panel->firstVisible; i < panel->firstVisible + panel->visibleCount; i++) {
-		changeImage(panel, i, i == panel->current);
+		if (i == panel->current)
+			continue;
+		dim = ((int) (uintptr_t) WMGetFromArray(panel->flags, i) & ICON_DIM);
+		changeImage(panel, i, 0, dim, True);
 	}
 }
 
@@ -249,68 +236,58 @@ static void scrollIcons(WSwitchPanel * panel, int delta)
  * 3 4 5
  * 6 7 8
  */
-static RImage *assemblePuzzleImage(RImage ** images, int width, int height)
+static RImage *assemblePuzzleImage(RImage **images, int width, int height)
 {
-	RImage *img = RCreateImage(width, height, 1);
+	RImage *img;
 	RImage *tmp;
 	int tw, th;
 	RColor color;
-	if (!img)
-		return NULL;
-
-	color.red = 0;
-	color.green = 0;
-	color.blue = 0;
-	color.alpha = 255;
-
-	RFillImage(img, &color);
 
 	tw = width - images[0]->width - images[2]->width;
 	th = height - images[0]->height - images[6]->height;
 
-	if (tw <= 0 || th <= 0) {
-		//XXX
+	if (tw <= 0 || th <= 0)
 		return NULL;
-	}
+
+	img = RCreateImage(width, height, 1);
+	if (!img)
+		return NULL;
+
+	color.red   =   0;
+	color.green =   0;
+	color.blue  =   0;
+	color.alpha = 255;
+	RFillImage(img, &color);
 
 	/* top */
-	if (tw > 0) {
-		tmp = RSmoothScaleImage(images[1], tw, images[1]->height);
-		RCopyArea(img, tmp, 0, 0, tmp->width, tmp->height, images[0]->width, 0);
-		RReleaseImage(tmp);
-	}
+	tmp = RSmoothScaleImage(images[1], tw, images[1]->height);
+	RCopyArea(img, tmp, 0, 0, tmp->width, tmp->height, images[0]->width, 0);
+	RReleaseImage(tmp);
+
 	/* bottom */
-	if (tw > 0) {
-		tmp = RSmoothScaleImage(images[7], tw, images[7]->height);
-		RCopyArea(img, tmp, 0, 0, tmp->width, tmp->height, images[6]->width, height - images[6]->height);
-		RReleaseImage(tmp);
-	}
+	tmp = RSmoothScaleImage(images[7], tw, images[7]->height);
+	RCopyArea(img, tmp, 0, 0, tmp->width, tmp->height, images[6]->width, height - images[6]->height);
+	RReleaseImage(tmp);
+
 	/* left */
-	if (th > 0) {
-		tmp = RSmoothScaleImage(images[3], images[3]->width, th);
-		RCopyArea(img, tmp, 0, 0, tmp->width, tmp->height, 0, images[0]->height);
-		RReleaseImage(tmp);
-	}
+	tmp = RSmoothScaleImage(images[3], images[3]->width, th);
+	RCopyArea(img, tmp, 0, 0, tmp->width, tmp->height, 0, images[0]->height);
+	RReleaseImage(tmp);
+
 	/* right */
-	if (th > 0) {
-		tmp = RSmoothScaleImage(images[5], images[5]->width, th);
-		RCopyArea(img, tmp, 0, 0, tmp->width, tmp->height, width - images[5]->width, images[2]->height);
-		RReleaseImage(tmp);
-	}
+	tmp = RSmoothScaleImage(images[5], images[5]->width, th);
+	RCopyArea(img, tmp, 0, 0, tmp->width, tmp->height, width - images[5]->width, images[2]->height);
+	RReleaseImage(tmp);
+
 	/* center */
-	if (tw > 0 && th > 0) {
-		tmp = RSmoothScaleImage(images[4], tw, th);
-		RCopyArea(img, tmp, 0, 0, tmp->width, tmp->height, images[0]->width, images[0]->height);
-		RReleaseImage(tmp);
-	}
+	tmp = RSmoothScaleImage(images[4], tw, th);
+	RCopyArea(img, tmp, 0, 0, tmp->width, tmp->height, images[0]->width, images[0]->height);
+	RReleaseImage(tmp);
 
 	/* corners */
 	RCopyArea(img, images[0], 0, 0, images[0]->width, images[0]->height, 0, 0);
-
 	RCopyArea(img, images[2], 0, 0, images[2]->width, images[2]->height, width - images[2]->width, 0);
-
 	RCopyArea(img, images[6], 0, 0, images[6]->width, images[6]->height, 0, height - images[6]->height);
-
 	RCopyArea(img, images[8], 0, 0, images[8]->width, images[8]->height,
 		  width - images[8]->width, height - images[8]->height);
 
@@ -336,7 +313,7 @@ static RImage *getTile(void)
 	return stile;
 }
 
-static void drawTitle(WSwitchPanel * panel, int idecks, char *title)
+static void drawTitle(WSwitchPanel *panel, int idecks, const char *title)
 {
 	char *ntitle;
 	int width = WMWidgetWidth(panel->win);
@@ -349,9 +326,9 @@ static void drawTitle(WSwitchPanel * panel, int idecks, char *title)
 
 	if (panel->bg) {
 		if (ntitle) {
-			if (strcmp(ntitle, title) != 0)
+			if (strcmp(ntitle, title) != 0) {
 				x = BORDER_SPACE;
-			else {
+			} else {
 				int w = WMWidthOfString(panel->font, ntitle, strlen(ntitle));
 
 				x = BORDER_SPACE + (idecks - panel->firstVisible) * ICON_TILE_SIZE +
@@ -362,6 +339,7 @@ static void drawTitle(WSwitchPanel * panel, int idecks, char *title)
 					x = width - BORDER_SPACE - w;
 			}
 		}
+
 		XClearWindow(dpy, WMWidgetXID(panel->win));
 		if (ntitle)
 			WMDrawString(panel->scr->wmscreen,
@@ -374,71 +352,54 @@ static void drawTitle(WSwitchPanel * panel, int idecks, char *title)
 		if (ntitle)
 			WMSetLabelText(panel->label, ntitle);
 	}
+
 	if (ntitle)
 		free(ntitle);
 }
 
-static WMArray *makeWindowListArray(WWindow *curwin, int include_unmapped, Bool class_only)
+static WMArray *makeWindowListArray(WScreen *scr, int include_unmapped, Bool class_only)
 {
 	WMArray *windows = WMCreateArray(10);
-	int fl;
-	WWindow *wwin;
+	WWindow *wwin = scr->focused_window;
 
-	for (fl = 0; fl < 2; fl++) {
-		for (wwin = curwin; wwin; wwin = wwin->prev) {
-			if (((!fl && canReceiveFocus(wwin) > 0) || (fl && canReceiveFocus(wwin) < 0)) &&
-			    (wwin->flags.mapped || include_unmapped)) {
-				if (class_only) {
-					if (!wwin->wm_class || !curwin->wm_class)
-						continue;
-					if (strcmp(wwin->wm_class, curwin->wm_class))
-						continue;
-				}
-
-				if (!WFLAGP(wwin, skip_switchpanel))
-					WMAddToArray(windows, wwin);
-			}
-		}
-		wwin = curwin;
-		/* start over from the beginning of the list */
-		while (wwin->next)
-			wwin = wwin->next;
-
-		for (wwin = curwin; wwin && wwin != curwin; wwin = wwin->prev) {
-			if (((!fl && canReceiveFocus(wwin) > 0) || (fl && canReceiveFocus(wwin) < 0)) &&
-			    (wwin->flags.mapped || include_unmapped)) {
-				if (class_only) {
-				    if (!wwin->wm_class || !curwin->wm_class)
+	while (wwin) {
+		if ((canReceiveFocus(wwin) != 0) &&
+		    (wwin->flags.mapped || wwin->flags.shaded || include_unmapped)) {
+			if (class_only)
+				if (!sameWindowClass(scr->focused_window, wwin))
 					continue;
-				    if (strcmp(wwin->wm_class, curwin->wm_class))
-					continue;
-				}
-
-				if (!WFLAGP(wwin, skip_switchpanel))
+			if (!WFLAGP(wwin, skip_switchpanel))
 				WMAddToArray(windows, wwin);
-			}
 		}
+		wwin = wwin->prev;
 	}
-
 	return windows;
 }
 
-WSwitchPanel *wInitSwitchPanel(WScreen * scr, WWindow * curwin, Bool class_only)
+static WMArray *makeWindowFlagsArray(int count)
+{
+	WMArray *flags = WMCreateArray(count);
+	int i;
+
+	for (i = 0; i < count; i++)
+		WMAddToArray(flags, (void *) 0);
+
+	return flags;
+}
+
+WSwitchPanel *wInitSwitchPanel(WScreen *scr, WWindow *curwin, Bool class_only)
 {
 	WWindow *wwin;
 	WSwitchPanel *panel = wmalloc(sizeof(WSwitchPanel));
 	WMFrame *viewport;
-	int i;
-	int width, height;
-	int iconsThatFitCount;
-	int count;
-	WMRect rect;
-	rect = wGetRectForHead(scr, wGetHeadForPointerLocation(scr));
+	int i, width, height, iconsThatFitCount, count;
+	WMRect rect = wGetRectForHead(scr, wGetHeadForPointerLocation(scr));
 
 	panel->scr = scr;
-
-	panel->windows = makeWindowListArray(curwin, wPreferences.swtileImage != NULL, class_only);
+	panel->windows = makeWindowListArray(scr, wPreferences.swtileImage != NULL, class_only);
 	count = WMGetArrayItemCount(panel->windows);
+	if (count)
+		panel->flags = makeWindowFlagsArray(count);
 
 	if (count == 0) {
 		WMFreeArray(panel->windows);
@@ -463,9 +424,9 @@ WSwitchPanel *wInitSwitchPanel(WScreen * scr, WWindow * curwin, Bool class_only)
 
 	panel->tileTmp = RCreateImage(ICON_TILE_SIZE, ICON_TILE_SIZE, 1);
 	panel->tile = getTile();
-	if (panel->tile && wPreferences.swbackImage[8]) {
+	if (panel->tile && wPreferences.swbackImage[8])
 		panel->bg = createBackImage(width + 2 * BORDER_SPACE, height + 2 * BORDER_SPACE);
-	}
+
 	if (!panel->tileTmp || !panel->tile) {
 		if (panel->bg)
 			RReleaseImage(panel->bg);
@@ -523,7 +484,7 @@ WSwitchPanel *wInitSwitchPanel(WScreen * scr, WWindow * curwin, Bool class_only)
 	WMRealizeWidget(panel->win);
 
 	WM_ITERATE_ARRAY(panel->windows, wwin, i) {
-		changeImage(panel, i, 0);
+		changeImage(panel, i, 0, False, True);
 	}
 
 	if (panel->bg) {
@@ -533,13 +494,13 @@ WSwitchPanel *wInitSwitchPanel(WScreen * scr, WWindow * curwin, Bool class_only)
 
 		XSetWindowBackgroundPixmap(dpy, WMWidgetXID(panel->win), pixmap);
 
-#ifdef SHAPE
-		if (mask && wShapeSupported)
+#ifdef USE_XSHAPE
+		if (mask && w_global.xext.shape.supported)
 			XShapeCombineMask(dpy, WMWidgetXID(panel->win), ShapeBounding, 0, 0, mask, ShapeSet);
 #endif
-
 		if (pixmap)
 			XFreePixmap(dpy, pixmap);
+
 		if (mask)
 			XFreePixmap(dpy, mask);
 	}
@@ -553,14 +514,14 @@ WSwitchPanel *wInitSwitchPanel(WScreen * scr, WWindow * curwin, Bool class_only)
 
 	panel->current = WMGetFirstInArray(panel->windows, curwin);
 	if (panel->current >= 0)
-		changeImage(panel, panel->current, 1);
+		changeImage(panel, panel->current, 1, False, False);
 
 	WMMapWidget(panel->win);
 
 	return panel;
 }
 
-void wSwitchPanelDestroy(WSwitchPanel * panel)
+void wSwitchPanelDestroy(WSwitchPanel *panel)
 {
 	int i;
 	RImage *image;
@@ -569,7 +530,7 @@ void wSwitchPanelDestroy(WSwitchPanel * panel)
 		Window info_win = panel->scr->info_window;
 		XEvent ev;
 		ev.xclient.type = ClientMessage;
-		ev.xclient.message_type = _XA_WM_IGNORE_FOCUS_EVENTS;
+		ev.xclient.message_type = w_global.atom.wm.ignore_focus_events;
 		ev.xclient.format = 32;
 		ev.xclient.data.l[0] = True;
 
@@ -587,78 +548,109 @@ void wSwitchPanelDestroy(WSwitchPanel * panel)
 		}
 		WMFreeArray(panel->images);
 	}
+
 	if (panel->win)
 		WMDestroyWidget(panel->win);
+
 	if (panel->icons)
 		WMFreeArray(panel->icons);
+
+	if (panel->flags)
+		WMFreeArray(panel->flags);
+
 	WMFreeArray(panel->windows);
-	if (panel->defIcon)
-		RReleaseImage(panel->defIcon);
+
 	if (panel->tile)
 		RReleaseImage(panel->tile);
+
 	if (panel->tileTmp)
 		RReleaseImage(panel->tileTmp);
+
 	if (panel->bg)
 		RReleaseImage(panel->bg);
+
 	if (panel->font)
 		WMReleaseFont(panel->font);
+
 	if (panel->white)
 		WMReleaseColor(panel->white);
+
 	wfree(panel);
 }
 
-WWindow *wSwitchPanelSelectNext(WSwitchPanel * panel, int back)
+WWindow *wSwitchPanelSelectNext(WSwitchPanel *panel, int back, int ignore_minimized, Bool class_only)
 {
-	WWindow *wwin;
+	WWindow *wwin, *curwin, *tmpwin;
 	int count = WMGetArrayItemCount(panel->windows);
+	int orig = panel->current;
+	int i;
+	Bool dim = False;
 
 	if (count == 0)
 		return NULL;
 
-	if (panel->win)
-		changeImage(panel, panel->current, 0);
+	if (!wPreferences.cycle_ignore_minimized)
+		ignore_minimized = False;
 
-	if (back)
-		panel->current--;
-	else
-		panel->current++;
+	if (ignore_minimized && canReceiveFocus(WMGetFromArray(panel->windows, (count + panel->current) % count)) < 0)
+		ignore_minimized = False;
 
-	wwin = WMGetFromArray(panel->windows, (count + panel->current) % count);
+	curwin = WMGetFromArray(panel->windows, orig);
+	do {
+		do {
+			if (back)
+				panel->current--;
+			else
+				panel->current++;
+
+			panel->current = (count + panel->current) % count;
+			wwin = WMGetFromArray(panel->windows, panel->current);
+
+			if (!class_only)
+				break;
+			if (panel->current == orig)
+				break;
+		} while (!sameWindowClass(wwin, curwin));
+	} while (ignore_minimized && panel->current != orig && canReceiveFocus(wwin) < 0);
+
+	WM_ITERATE_ARRAY(panel->windows, tmpwin, i) {
+		if (i == panel->current)
+			continue;
+		if (!class_only || sameWindowClass(tmpwin, curwin))
+			changeImage(panel, i, 0, False, False);
+		else {
+			if (i == orig)
+				dim = True;
+			changeImage(panel, i, 0, True, False);
+		}
 
-	if (back) {
-		if (panel->current < 0)
-			scrollIcons(panel, count);
-		else if (panel->current < panel->firstVisible)
-			scrollIcons(panel, -1);
-	} else {
-		if (panel->current >= count)
-			scrollIcons(panel, -count);
-		else if (panel->current - panel->firstVisible >= panel->visibleCount)
-			scrollIcons(panel, 1);
 	}
 
-	panel->current = (count + panel->current) % count;
+	if (panel->current < panel->firstVisible)
+		scrollIcons(panel, panel->current - panel->firstVisible);
+	else if (panel->current - panel->firstVisible >= panel->visibleCount)
+		scrollIcons(panel, panel->current - panel->firstVisible - panel->visibleCount + 1);
 
 	if (panel->win) {
 		drawTitle(panel, panel->current, wwin->frame->title);
-
-		changeImage(panel, panel->current, 1);
+		if (panel->current != orig)
+			changeImage(panel, orig, 0, dim, False);
+		changeImage(panel, panel->current, 1, False, False);
 	}
 
 	return wwin;
 }
 
-WWindow *wSwitchPanelSelectFirst(WSwitchPanel * panel, int back)
+WWindow *wSwitchPanelSelectFirst(WSwitchPanel *panel, int back)
 {
 	WWindow *wwin;
 	int count = WMGetArrayItemCount(panel->windows);
+	char *title;
+	int i;
 
 	if (count == 0)
 		return NULL;
 
-	if (panel->win)
-		changeImage(panel, panel->current, 0);
-
 	if (back) {
 		panel->current = count - 1;
 		scrollIcons(panel, count);
@@ -668,16 +660,19 @@ WWindow *wSwitchPanelSelectFirst(WSwitchPanel * panel, int back)
 	}
 
 	wwin = WMGetFromArray(panel->windows, panel->current);
+	title = wwin->frame->title;
 
 	if (panel->win) {
-		drawTitle(panel, panel->current, wwin->frame->title);
+		for (WMArrayFirst(panel->windows, &(i)); (i) != WANotFound; WMArrayNext(panel->windows, &(i)))
+			changeImage(panel, i, i == panel->current, False, False);
 
-		changeImage(panel, panel->current, 1);
+		drawTitle(panel, panel->current, title);
 	}
+
 	return wwin;
 }
 
-WWindow *wSwitchPanelHandleEvent(WSwitchPanel * panel, XEvent * event)
+WWindow *wSwitchPanelHandleEvent(WSwitchPanel *panel, XEvent *event)
 {
 	WMFrame *icon;
 	int i;
@@ -686,11 +681,7 @@ WWindow *wSwitchPanelHandleEvent(WSwitchPanel * panel, XEvent * event)
 	if (!panel->win)
 		return NULL;
 
-	/*   if (event->type == LeaveNotify) {
-	   if (event->xcrossing.window == WMWidgetXID(panel->win))
-	   focus= 0;
-	   } else */ if (event->type == MotionNotify) {
-
+	if (event->type == MotionNotify) {
 		WM_ITERATE_ARRAY(panel->icons, icon, i) {
 			if (WMWidgetXID(icon) == event->xmotion.window) {
 				focus = i;
@@ -698,11 +689,13 @@ WWindow *wSwitchPanelHandleEvent(WSwitchPanel * panel, XEvent * event)
 			}
 		}
 	}
+
 	if (focus >= 0 && panel->current != focus) {
 		WWindow *wwin;
 
-		changeImage(panel, panel->current, 0);
-		changeImage(panel, focus, 1);
+		WM_ITERATE_ARRAY(panel->windows, wwin, i) {
+			changeImage(panel, i, i == focus, False, False);
+		}
 		panel->current = focus;
 
 		wwin = WMGetFromArray(panel->windows, focus);
@@ -715,9 +708,10 @@ WWindow *wSwitchPanelHandleEvent(WSwitchPanel * panel, XEvent * event)
 	return NULL;
 }
 
-Window wSwitchPanelGetWindow(WSwitchPanel * swpanel)
+Window wSwitchPanelGetWindow(WSwitchPanel *swpanel)
 {
 	if (!swpanel->win)
 		return None;
+
 	return WMWidgetXID(swpanel->win);
 }
