@@ -23,7 +23,7 @@
 
 #include <X11/Xlib.h>
 #include <X11/Xutil.h>
-#ifdef SHAPE
+#ifdef USE_XSHAPE
 #include <X11/extensions/shape.h>
 #endif
 #ifdef KEEP_XKB_LOCK_STATUS
@@ -36,7 +36,7 @@
 #include <math.h>
 
 /* For getting mouse wheel mappings from WINGs */
-#include <WINGs/WINGsP.h>
+#include <WINGs/WINGs.h>
 
 #include "WindowMaker.h"
 #include "GNUstep.h"
@@ -49,7 +49,7 @@
 #include "properties.h"
 #include "actions.h"
 #include "client.h"
-#include "funcs.h"
+#include "colormap.h"
 #include "keybind.h"
 #include "stacking.h"
 #include "defaults.h"
@@ -57,33 +57,21 @@
 #include "xinerama.h"
 #include "appmenu.h"
 #include "appicon.h"
+#include "superfluous.h"
+#include "rootmenu.h"
+#include "placement.h"
+#include "misc.h"
+#include "startup.h"
+#include "winmenu.h"
+#include "osdep.h"
 
-#ifdef MWM_HINTS
+#ifdef USE_MWM_HINTS
 # include "motif.h"
 #endif
 #include "wmspec.h"
 
 #define MOD_MASK wPreferences.modifier_mask
 
-/****** Global Variables ******/
-extern WShortKey wKeyBindings[WKBD_LAST];
-
-#ifdef SHAPE
-extern Bool wShapeSupported;
-#endif
-
-/* contexts */
-extern XContext wWinContext;
-
-/* protocol atoms */
-extern Atom _XA_WM_DELETE_WINDOW;
-extern Atom _XA_GNUSTEP_WM_MINIATURIZE_WINDOW;
-extern Atom _XA_WINDOWMAKER_STATE;
-extern WPreferences wPreferences;
-extern Time LastTimestamp;
-
-/* superfluous... */
-extern void DoWindowBirth(WWindow *wwin);
 
 /***** Local Stuff *****/
 static WWindowState *windowState = NULL;
@@ -109,6 +97,8 @@ static void titlebarMouseDown(WCoreWindow *sender, void *data, XEvent *event);
 static void titlebarDblClick(WCoreWindow *sender, void *data, XEvent *event);
 static void resizebarMouseDown(WCoreWindow *sender, void *data, XEvent *event);
 
+static void release_wwindowstate(WWindowState *wstate);
+
 /****** Notification Observers ******/
 
 static void appearanceObserver(void *self, WMNotification * notif)
@@ -139,6 +129,7 @@ static void appearanceObserver(void *self, WMNotification * notif)
 }
 
 
+/* Return the WWindow associated with a given (Xlib) Window. */
 WWindow *wWindowFor(Window window)
 {
 	WObjDescriptor *desc;
@@ -146,7 +137,7 @@ WWindow *wWindowFor(Window window)
 	if (window == None)
 		return NULL;
 
-	if (XFindContext(dpy, window, wWinContext, (XPointer *) & desc) == XCNOENT)
+	if (XFindContext(dpy, window, w_global.context.client_win, (XPointer *) & desc) == XCNOENT)
 		return NULL;
 
 	if (desc->parent_type == WCLASS_WINDOW)
@@ -226,7 +217,7 @@ void wWindowDestroy(WWindow *wwin)
 	if (wwin->cmap_windows)
 		XFree(wwin->cmap_windows);
 
-	XDeleteContext(dpy, wwin->client_win, wWinContext);
+	XDeleteContext(dpy, wwin->client_win, w_global.context.client_win);
 
 	if (wwin->frame)
 		wFrameWindowDestroy(wwin->frame);
@@ -297,12 +288,39 @@ static void setupGNUstepHints(WWindow *wwin, GNUstepWMAttributes *gs_hints)
 		wwin->client_flags.no_appicon = 1;
 }
 
+static void discard_hints_from_gtk(WWindow *wwin)
+{
+	Atom type;
+	int format;
+	unsigned long nb_item, nb_remain;
+	unsigned char *result;
+	int status;
+
+	status = XGetWindowProperty(dpy, wwin->client_win, w_global.atom.desktop.gtk_object_path, 0, 16, False,
+	                            AnyPropertyType, &type, &format, &nb_item, &nb_remain, &result);
+	if (status != Success)
+		return;
+
+	/* If we're here, that means the Property exists. We don't care what is inside, it means it is a GTK-based application */
+
+	if (result != NULL)
+		XFree(result);
+
+	/* GTK is asking to remove these decorations: */
+	wwin->client_flags.no_titlebar = 0;
+	wwin->client_flags.no_close_button = 0;
+	wwin->client_flags.no_miniaturize_button = 0;
+	wwin->client_flags.no_resizebar = 0;
+}
+
 void wWindowSetupInitialAttributes(WWindow *wwin, int *level, int *workspace)
 {
 	WScreen *scr = wwin->screen_ptr;
 
 	/* sets global default stuff */
-	wDefaultFillAttributes(scr, wwin->wm_instance, wwin->wm_class, &wwin->client_flags, NULL, True);
+	wDefaultFillAttributes(wwin->wm_instance, wwin->wm_class, &wwin->user_flags, NULL, True);
+	wwin->defined_user_flags = wwin->user_flags;
+
 	/*
 	 * Decoration setting is done in this precedence (lower to higher)
 	 * - use global default in the resource database
@@ -311,25 +329,25 @@ void wWindowSetupInitialAttributes(WWindow *wwin, int *level, int *workspace)
 	 * - set hints specified for the app in the resource DB
 	 *
 	 */
-	WSETUFLAG(wwin, broken_close, 0);
+	wwin->client_flags.broken_close = 0;
 
 	if (wwin->protocols.DELETE_WINDOW)
-		WSETUFLAG(wwin, kill_close, 0);
+		wwin->client_flags.kill_close = 0;
 	else
-		WSETUFLAG(wwin, kill_close, 1);
+		wwin->client_flags.kill_close = 1;
 
 	/* transients can't be iconified or maximized */
 	if (wwin->transient_for != None && wwin->transient_for != scr->root_win) {
-		WSETUFLAG(wwin, no_miniaturizable, 1);
-		WSETUFLAG(wwin, no_miniaturize_button, 1);
+		wwin->client_flags.no_miniaturizable = 1;
+		wwin->client_flags.no_miniaturize_button = 1;
 	}
 
 	/* if the window can't be resized, remove the resizebar */
 	if (wwin->normal_hints->flags & (PMinSize | PMaxSize)
 	    && (wwin->normal_hints->min_width == wwin->normal_hints->max_width)
 	    && (wwin->normal_hints->min_height == wwin->normal_hints->max_height)) {
-		WSETUFLAG(wwin, no_resizable, 1);
-		WSETUFLAG(wwin, no_resizebar, 1);
+		wwin->client_flags.no_resizable = 1;
+		wwin->client_flags.no_resizebar = 1;
 	}
 
 	/* set GNUstep window attributes */
@@ -351,16 +369,15 @@ void wWindowSetupInitialAttributes(WWindow *wwin, int *level, int *workspace)
 	} else {
 		int tmp_workspace = -1;
 		int tmp_level = INT_MIN;	/* INT_MIN is never used by the window levels */
-		Bool check;
 
-		check = False;
-
-#ifdef MWM_HINTS
+#ifdef USE_MWM_HINTS
 		wMWMCheckClientHints(wwin);
-#endif				/* MWM_HINTS */
+#endif	/* USE_MWM_HINTS */
 
-		if (!check)
-			check = wNETWMCheckClientHints(wwin, &tmp_level, &tmp_workspace);
+		wNETWMCheckClientHints(wwin, &tmp_level, &tmp_workspace);
+
+		if (wPreferences.ignore_gtk_decoration_hints)
+			discard_hints_from_gtk(wwin);
 
 		/* window levels are between INT_MIN+1 and INT_MAX, so if we still
 		 * have INT_MIN that means that no window level was requested. -Dan
@@ -393,8 +410,8 @@ void wWindowSetupInitialAttributes(WWindow *wwin, int *level, int *workspace)
 	 * Set attributes specified only for that window/class.
 	 * This might do duplicate work with the 1st wDefaultFillAttributes().
 	 */
-	wDefaultFillAttributes(scr, wwin->wm_instance, wwin->wm_class,
-			       &wwin->user_flags, &wwin->defined_user_flags, False);
+	wDefaultFillAttributes(wwin->wm_instance, wwin->wm_class, &wwin->user_flags,
+			       &wwin->defined_user_flags, False);
 	/*
 	 * Sanity checks for attributes that depend on other attributes
 	 */
@@ -414,25 +431,15 @@ void wWindowSetupInitialAttributes(WWindow *wwin, int *level, int *workspace)
 	    && wwin->user_flags.floating && wwin->defined_user_flags.floating)
 		wwin->user_flags.sunken = 0;
 
-	WSETUFLAG(wwin, no_shadeable, WFLAGP(wwin, no_titlebar));
+	/* A window that does not have a title cannot be Shaded and we don't let user override this */
+	wwin->client_flags.no_shadeable = WFLAGP(wwin, no_titlebar);
+	wwin->defined_user_flags.no_shadeable = 0;
 
 	/* windows that have takefocus=False shouldn't take focus at all */
 	if (wwin->focus_mode == WFM_NO_INPUT)
 		wwin->client_flags.no_focusable = 1;
 }
 
-Bool wWindowCanReceiveFocus(WWindow *wwin)
-{
-	if (!wwin->flags.mapped && (!wwin->flags.shaded || wwin->flags.hidden))
-		return False;
-	if (WFLAGP(wwin, no_focusable) || wwin->flags.miniaturized)
-		return False;
-	if (wwin->frame->workspace != wwin->screen_ptr->current_workspace)
-		return False;
-
-	return True;
-}
-
 Bool wWindowObscuresWindow(WWindow *wwin, WWindow *obscured)
 {
 	int w1, h1, w2, h2;
@@ -608,6 +615,7 @@ WWindow *wManageWindow(WScreen *scr, Window window)
 	/* mutex. */
 	XGrabServer(dpy);
 	XSync(dpy, False);
+
 	/* make sure the window is still there */
 	if (!XGetWindowAttributes(dpy, window, &wattribs)) {
 		XUngrabServer(dpy);
@@ -633,13 +641,13 @@ WWindow *wManageWindow(WScreen *scr, Window window)
 	title = wNETWMGetWindowName(window);
 	if (title)
 		wwin->flags.net_has_title = 1;
-	if (!title && !wFetchName(dpy, window, &title))
+	else if (!wFetchName(dpy, window, &title))
 		title = NULL;
 
-	XSaveContext(dpy, window, wWinContext, (XPointer) & wwin->client_descriptor);
+	XSaveContext(dpy, window, w_global.context.client_win, (XPointer) & wwin->client_descriptor);
 
-#ifdef SHAPE
-	if (wShapeSupported) {
+#ifdef USE_XSHAPE
+	if (w_global.xext.shape.supported) {
 		int junk;
 		unsigned int ujunk;
 		int b_shaped;
@@ -651,9 +659,7 @@ WWindow *wManageWindow(WScreen *scr, Window window)
 	}
 #endif
 
-	/*
-	 * Get hints and other information in properties
-	 */
+	/* Get hints and other information in properties */
 	PropGetWMClass(window, &wwin->wm_class, &wwin->wm_instance);
 
 	/* setup descriptor */
@@ -688,7 +694,6 @@ WWindow *wManageWindow(WScreen *scr, Window window)
 
 	if (wwin->wm_hints) {
 		if (wwin->wm_hints->flags & StateHint) {
-
 			if (wwin->wm_hints->initial_state == IconicState) {
 				wwin->flags.miniaturized = 1;
 			} else if (wwin->wm_hints->initial_state == WithdrawnState) {
@@ -751,23 +756,22 @@ WWindow *wManageWindow(WScreen *scr, Window window)
 	wwin->orig_main_window = wwin->main_window;
 
 	if (wwin->flags.is_gnustep)
-		WSETUFLAG(wwin, shared_appicon, 0);
+		wwin->client_flags.shared_appicon = 0;
 
 	if (wwin->main_window) {
-		extern Atom _XA_WINDOWMAKER_MENU;
 		XTextProperty text_prop;
 
-		if (XGetTextProperty(dpy, wwin->main_window, &text_prop, _XA_WINDOWMAKER_MENU))
-			WSETUFLAG(wwin, shared_appicon, 0);
+		if (XGetTextProperty(dpy, wwin->main_window, &text_prop, w_global.atom.wmaker.menu))
+			wwin->client_flags.shared_appicon = 0;
 	}
 
 	if (wwin->flags.is_dockapp)
-		WSETUFLAG(wwin, shared_appicon, 0);
+		wwin->client_flags.shared_appicon = 0;
 
 	if (wwin->main_window) {
             WApplication *app = wApplicationOf(wwin->main_window);
             if (app && app->app_icon)
-		WSETUFLAG(wwin, shared_appicon, 0);
+		wwin->client_flags.shared_appicon = 0;
         }
 
 	if (!withdraw && wwin->main_window && WFLAGP(wwin, shared_appicon)) {
@@ -796,7 +800,6 @@ WWindow *wManageWindow(WScreen *scr, Window window)
 				}
 			}
 			wwin->fake_group = fPtr;
-			/*wwin->group_id = fPtr->leader; */
 			wwin->main_window = fPtr->leader;
 			wfree(buffer);
 		} else {
@@ -814,11 +817,12 @@ WWindow *wManageWindow(WScreen *scr, Window window)
 				fPtr->origLeader = wwin->main_window;
 			}
 			wwin->fake_group = fPtr;
-			/*wwin->group_id = fPtr->leader; */
 			wwin->main_window = fPtr->leader;
 		}
+
 		if (instance)
 			free(instance);
+
 		if (class)
 			free(class);
 #undef ADEQUATE
@@ -837,11 +841,10 @@ WWindow *wManageWindow(WScreen *scr, Window window)
 
 	/* apply previous state if it exists and we're in startup */
 	if (scr->flags.startup && wm_state >= 0) {
-		if (wm_state == IconicState) {
+		if (wm_state == IconicState)
 			wwin->flags.miniaturized = 1;
-		} else if (wm_state == WithdrawnState) {
+		else if (wm_state == WithdrawnState)
 			withdraw = True;
-		}
 	}
 
 	/* if there is a saved state (from file), restore it */
@@ -921,6 +924,7 @@ WWindow *wManageWindow(WScreen *scr, Window window)
 				}
 			}
 		}
+
 		if (wstate != NULL)
 			wfree(wstate);
 	}
@@ -956,6 +960,7 @@ WWindow *wManageWindow(WScreen *scr, Window window)
 		width = win_state->state->w;
 		height = win_state->state->h;
 	}
+
 	wWindowConstrainSize(wwin, &width, &height);
 
 	/* do not ask for window placement if the window is
@@ -984,9 +989,7 @@ WWindow *wManageWindow(WScreen *scr, Window window)
 				y = transientOwner->frame_y +
 				    abs((transientOwner->frame->core->height - height) / 3) + offs;
 
-				/*
-				 * limit transient windows to be inside their parent's head
-				 */
+				/* limit transient windows to be inside their parent's head */
 				rect.pos.x = transientOwner->frame_x;
 				rect.pos.y = transientOwner->frame_y;
 				rect.size.width = transientOwner->frame->core->width;
@@ -1008,28 +1011,25 @@ WWindow *wManageWindow(WScreen *scr, Window window)
 			} else {
 				PlaceWindow(wwin, &x, &y, width, height);
 			}
-			if (wPreferences.window_placement == WPM_MANUAL) {
+
+			if (wPreferences.window_placement == WPM_MANUAL)
 				dontBring = True;
-			}
+
 		} else if (scr->xine_info.count && (wwin->normal_hints->flags & PPosition)) {
 			int head, flags;
 			WMRect rect;
 			int reposition = 0;
 
-			/*
-			 * Make spash screens come out in the center of a head
+			/* Make spash screens come out in the center of a head
 			 * trouble is that most splashies never get here
 			 * they are managed trough atoms but god knows where.
 			 * Dan, do you know ? -peter
 			 *
 			 * Most of them are not managed, they have set
 			 * OverrideRedirect, which means we can't do anything about
-			 * them. -alfredo
-			 */
+			 * them. -alfredo */
 			{
-				/*
-				 * xinerama checks for: across head and dead space
-				 */
+				/* xinerama checks for: across head and dead space */
 				rect.pos.x = x;
 				rect.pos.y = y;
 				rect.size.width = width;
@@ -1084,9 +1084,7 @@ WWindow *wManageWindow(WScreen *scr, Window window)
 			wwin->flags.omnipresent ^= 1;
 	}
 
-	/*
-	 * Create frame, borders and do reparenting
-	 */
+	/* Create frame, borders and do reparenting */
 	foo = WFF_LEFT_BUTTON | WFF_RIGHT_BUTTON;
 #ifdef XKB_BUTTON_HINT
 	if (wPreferences.modelock)
@@ -1094,8 +1092,10 @@ WWindow *wManageWindow(WScreen *scr, Window window)
 #endif
 	if (HAS_TITLEBAR(wwin))
 		foo |= WFF_TITLEBAR;
+
 	if (HAS_RESIZEBAR(wwin))
 		foo |= WFF_RESIZEBAR;
+
 	if (HAS_BORDER(wwin))
 		foo |= WFF_BORDER;
 
@@ -1106,11 +1106,14 @@ WWindow *wManageWindow(WScreen *scr, Window window)
 					 &wPreferences.window_title_max_height,
 					 foo,
 					 scr->window_title_texture,
-					 scr->resizebar_texture, scr->window_title_color, &scr->title_font);
+					 scr->resizebar_texture, scr->window_title_color, &scr->title_font,
+					 wattribs.depth, wattribs.visual, wattribs.colormap);
 
 	wwin->frame->flags.is_client_window_frame = 1;
 	wwin->frame->flags.justification = wPreferences.title_justification;
 
+	wNETWMCheckInitialFrameState(wwin);
+
 	/* setup button images */
 	wWindowUpdateButtonImages(wwin);
 
@@ -1118,8 +1121,10 @@ WWindow *wManageWindow(WScreen *scr, Window window)
 	foo = 0;
 	if (WFLAGP(wwin, no_close_button))
 		foo |= WFF_RIGHT_BUTTON;
+
 	if (WFLAGP(wwin, no_miniaturize_button))
 		foo |= WFF_LEFT_BUTTON;
+
 #ifdef XKB_BUTTON_HINT
 	if (WFLAGP(wwin, no_language_button) || WFLAGP(wwin, no_focusable))
 		foo |= WFF_LANGUAGE_BUTTON;
@@ -1159,17 +1164,13 @@ WWindow *wManageWindow(WScreen *scr, Window window)
 	/*
 	 * wWindowConfigure() will init the client window's size
 	 * (wwin->client.{width,height}) and all other geometry
-	 * related variables (frame_x,frame_y)
-	 */
+	 * related variables (frame_x,frame_y) */
 	wWindowConfigure(wwin, x, y, width, height);
 
 	/* to make sure the window receives it's new position after reparenting */
 	wWindowSynthConfigureNotify(wwin);
 
-	/*
-	 * Setup descriptors and save window to internal
-	 * lists
-	 */
+	/* Setup descriptors and save window to internal lists */
 	if (wwin->main_window != None) {
 		WApplication *app;
 		WWindow *leader;
@@ -1185,10 +1186,8 @@ WWindow *wManageWindow(WScreen *scr, Window window)
 		if (app) {
 			app->last_workspace = workspace;
 
-			/*
-			 * Do application specific stuff, like setting application
-			 * wide attributes.
-			 */
+			/* Do application specific stuff, like setting application
+			 * wide attributes. */
 
 			if (wwin->flags.hidden) {
 				/* if the window was set to hidden because it was hidden
@@ -1222,8 +1221,7 @@ WWindow *wManageWindow(WScreen *scr, Window window)
 
 		/* The following "if" is to avoid crashing of clients that expect
 		 * WM_STATE set before they get mapped. Else WM_STATE is set later,
-		 * after the return from this function.
-		 */
+		 * after the return from this function. */
 		if (wwin->wm_hints && (wwin->wm_hints->flags & StateHint))
 			wClientSetState(wwin, wwin->wm_hints->initial_state, None);
 		else
@@ -1271,6 +1269,7 @@ WWindow *wManageWindow(WScreen *scr, Window window)
 		tmp = scr->focused_window;
 		while (tmp->prev)
 			tmp = tmp->prev;
+
 		tmp->prev = wwin;
 		wwin->next = tmp;
 		wwin->prev = NULL;
@@ -1288,9 +1287,7 @@ WWindow *wManageWindow(WScreen *scr, Window window)
 
 	XUngrabServer(dpy);
 
-	/*
-	 * Final preparations before window is ready to go
-	 */
+	/* Final preparations before window is ready to go */
 	wFrameWindowChangeState(wwin->frame, WS_UNFOCUSED);
 
 	if (!wwin->flags.miniaturized && workspace == scr->current_workspace && !wwin->flags.hidden) {
@@ -1344,7 +1341,7 @@ WWindow *wManageWindow(WScreen *scr, Window window)
 }
 
 WWindow *wManageInternalWindow(WScreen *scr, Window window, Window owner,
-			       char *title, int x, int y, int width, int height)
+			       const char *title, int x, int y, int width, int height)
 {
 	WWindow *wwin;
 	int foo;
@@ -1355,10 +1352,10 @@ WWindow *wManageInternalWindow(WScreen *scr, Window window, Window owner,
 
 	wwin->flags.internal_window = 1;
 
-	WSETUFLAG(wwin, omnipresent, 1);
-	WSETUFLAG(wwin, no_shadeable, 1);
-	WSETUFLAG(wwin, no_resizable, 1);
-	WSETUFLAG(wwin, no_miniaturizable, 1);
+	wwin->client_flags.omnipresent = 1;
+	wwin->client_flags.no_shadeable = 1;
+	wwin->client_flags.no_resizable = 1;
+	wwin->client_flags.no_miniaturizable = 1;
 
 	wwin->focus_mode = WFM_PASSIVE;
 	wwin->client_win = window;
@@ -1385,9 +1382,10 @@ WWindow *wManageInternalWindow(WScreen *scr, Window window, Window owner,
 					 &wPreferences.window_title_max_height,
 					 foo,
 					 scr->window_title_texture,
-					 scr->resizebar_texture, scr->window_title_color, &scr->title_font);
+					 scr->resizebar_texture, scr->window_title_color, &scr->title_font,
+					 scr->w_depth, scr->w_visual, scr->w_colormap);
 
-	XSaveContext(dpy, window, wWinContext, (XPointer) & wwin->client_descriptor);
+	XSaveContext(dpy, window, w_global.context.client_win, (XPointer) & wwin->client_descriptor);
 
 	wwin->frame->flags.is_client_window_frame = 1;
 	wwin->frame->flags.justification = wPreferences.title_justification;
@@ -1492,6 +1490,11 @@ void wUnmanageWindow(WWindow *wwin, Bool restore, Bool destroyed)
 	if (wwin->flags.menu_open_for_me)
 		CloseWindowMenu(scr);
 
+	/* Don't restore focus to this window after a window exits
+	 * fullscreen mode */
+	if (scr->bfs_focused_window == wwin)
+		scr->bfs_focused_window = NULL;
+
 	if (!destroyed) {
 		if (!wwin->flags.internal_window)
 			XRemoveFromSaveSet(dpy, wwin->client_win);
@@ -1624,9 +1627,12 @@ void wUnmanageWindow(WWindow *wwin, Bool restore, Bool destroyed)
 	WApplication *napp = scr->focused_window ? wApplicationOf(scr->focused_window->main_window) : NULL;
 	if (oapp && oapp != napp) {
 		wAppMenuUnmap(oapp->menu);
-		wApplicationDeactivate(oapp);
+		if (wPreferences.highlight_active_app)
+			wApplicationDeactivate(oapp);
 	}
 
+	wNETCleanupFrameExtents(wwin);
+
 	wWindowDestroy(wwin);
 	XFlush(dpy);
 }
@@ -1656,6 +1662,61 @@ void wWindowUnmap(WWindow *wwin)
 	XUnmapWindow(dpy, wwin->frame->core->window);
 }
 
+void wWindowSingleFocus(WWindow *wwin)
+{
+	WScreen *scr;
+	int x, y, move = 0;
+
+	if (!wwin)
+		return;
+
+	scr = wwin->screen_ptr;
+	wMakeWindowVisible(wwin);
+
+	x = wwin->frame_x;
+	y = wwin->frame_y;
+
+	/* bring window back to visible area */
+	move = wScreenBringInside(scr, &x, &y, wwin->frame->core->width, wwin->frame->core->height);
+
+	if (move) {
+		wWindowConfigure(wwin, x, y, wwin->client.width, wwin->client.height);
+	}
+}
+
+void wWindowFocusPrev(WWindow *wwin, Bool inSameWorkspace)
+{
+	WWindow *tmp;
+
+	if (!wwin || !wwin->prev)
+		return;
+
+	tmp = wwin;
+	while (tmp->prev)
+		tmp = tmp->prev;
+
+	if (inSameWorkspace)
+		while (tmp && (tmp->frame->workspace != wwin->frame->workspace))
+			tmp = tmp->next;
+
+	wWindowSingleFocus(tmp);
+}
+
+void wWindowFocusNext(WWindow *wwin, Bool inSameWorkspace)
+{
+	WWindow *tmp;
+
+	if (!wwin || !wwin->prev)
+		return;
+
+	tmp = wwin->prev;
+	if (inSameWorkspace)
+		while (tmp && (tmp->frame->workspace != wwin->frame->workspace))
+			tmp = tmp->prev;
+
+	wWindowSingleFocus(tmp);
+}
+
 void wWindowFocus(WWindow *wwin, WWindow *owin)
 {
 	WWindow *nowner;
@@ -1741,15 +1802,13 @@ void wWindowUnfocus(WWindow *wwin)
 	WMPostNotificationName(WMNChangedFocus, wwin, (void *)False);
 }
 
-void wWindowUpdateName(WWindow *wwin, char *newTitle)
+void wWindowUpdateName(WWindow *wwin, const char *newTitle)
 {
-	char *title;
+	const char *title;
 
 	if (!wwin->frame)
 		return;
 
-	wwin->flags.wm_name_changed = 1;
-
 	if (!newTitle)
 		title = DEF_WINDOW_TITLE; /* the hint was removed */
 	else
@@ -1939,6 +1998,30 @@ void wWindowChangeWorkspace(WWindow *wwin, int workspace)
 		wWindowUnmap(wwin);
 }
 
+void wWindowChangeWorkspaceRelative(WWindow *wwin, int amount)
+{
+	WScreen *scr = wwin->screen_ptr;
+	int w = scr->current_workspace + amount;
+
+	if (amount < 0) {
+		if (w >= 0) {
+			wWindowChangeWorkspace(wwin, w);
+		} else if (wPreferences.ws_cycle) {
+			wWindowChangeWorkspace(wwin, scr->workspace_count + w);
+		}
+	} else if (amount > 0) {
+		if (w < scr->workspace_count) {
+			wWindowChangeWorkspace(wwin, w);
+		} else if (wPreferences.ws_advance) {
+			int workspace = WMIN(w, MAX_WORKSPACES - 1);
+			wWorkspaceMake(scr, workspace);
+			wWindowChangeWorkspace(wwin, workspace);
+		} else if (wPreferences.ws_cycle) {
+			wWindowChangeWorkspace(wwin, w % scr->workspace_count);
+		}
+	}
+}
+
 void wWindowSynthConfigureNotify(WWindow *wwin)
 {
 	XEvent sevent;
@@ -2043,16 +2126,19 @@ void wWindowConfigure(WWindow *wwin, int req_x, int req_y, int req_width, int re
 	wwin->frame_x = req_x;
 	wwin->frame_y = req_y;
 	if (HAS_BORDER(wwin)) {
-		wwin->client.x += FRAME_BORDER_WIDTH;
-		wwin->client.y += FRAME_BORDER_WIDTH;
+		wwin->client.x += wwin->screen_ptr->frame_border_width;
+		wwin->client.y += wwin->screen_ptr->frame_border_width;
 	}
-#ifdef SHAPE
-	if (wShapeSupported && wwin->flags.shaped && resize)
+#ifdef USE_XSHAPE
+	if (w_global.xext.shape.supported && wwin->flags.shaped && resize)
 		wWindowSetShape(wwin);
 #endif
 
 	if (synth_notify)
 		wWindowSynthConfigureNotify(wwin);
+
+	wNETFrameExtents(wwin);
+
 	XFlush(dpy);
 }
 
@@ -2077,8 +2163,8 @@ void wWindowMove(WWindow *wwin, int req_x, int req_y)
 	wwin->client.x = req_x;
 	wwin->client.y = req_y + wwin->frame->top_width;
 	if (HAS_BORDER(wwin)) {
-		wwin->client.x += FRAME_BORDER_WIDTH;
-		wwin->client.y += FRAME_BORDER_WIDTH;
+		wwin->client.x += wwin->screen_ptr->frame_border_width;
+		wwin->client.y += wwin->screen_ptr->frame_border_width;
 	}
 
 	XMoveWindow(dpy, wwin->frame->core->window, req_x, req_y);
@@ -2118,7 +2204,7 @@ void wWindowUpdateButtonImages(WWindow *wwin)
 			}
 
 			if (!fwin->lbutton_image) {
-				fwin->lbutton_image = wPixmapCreate(scr, pixmap, mask);
+				fwin->lbutton_image = wPixmapCreate(pixmap, mask);
 				fwin->lbutton_image->client_owned = 1;
 				fwin->lbutton_image->client_owned_mask = 1;
 			}
@@ -2159,7 +2245,7 @@ void wWindowUpdateButtonImages(WWindow *wwin)
 			}
 
 			if (!fwin->rbutton_image) {
-				fwin->rbutton_image = wPixmapCreate(scr, pixmap, mask);
+				fwin->rbutton_image = wPixmapCreate(pixmap, mask);
 				fwin->rbutton_image->client_owned = 1;
 				fwin->rbutton_image->client_owned_mask = 1;
 			}
@@ -2217,6 +2303,8 @@ void wWindowConfigureBorders(WWindow *wwin)
 			flags |= WFF_BORDER;
 		if (wwin->flags.shaded)
 			flags |= WFF_IS_SHADED;
+		if (wwin->flags.selected)
+			flags |= WFF_SELECTED;
 
 		oldh = wwin->frame->top_width;
 		wFrameWindowUpdateBorders(wwin->frame, flags);
@@ -2265,14 +2353,14 @@ void wWindowConfigureBorders(WWindow *wwin)
 		if (flags != 0)
 			wFrameWindowHideButton(wwin->frame, flags);
 
-#ifdef SHAPE
-		if (wShapeSupported && wwin->flags.shaped)
+#ifdef USE_XSHAPE
+		if (w_global.xext.shape.supported && wwin->flags.shaped)
 			wWindowSetShape(wwin);
 #endif
 	}
 }
 
-void wWindowSaveState(WWindow * wwin)
+void wWindowSaveState(WWindow *wwin)
 {
 	long data[10];
 	int i;
@@ -2300,8 +2388,9 @@ void wWindowSaveState(WWindow * wwin)
 		    WMCountInArray(wwin->screen_ptr->shortcutWindows[i], wwin))
 			data[9] |= 1 << i;
 	}
-	XChangeProperty(dpy, wwin->client_win, _XA_WINDOWMAKER_STATE,
-			_XA_WINDOWMAKER_STATE, 32, PropModeReplace, (unsigned char *)data, 10);
+
+	XChangeProperty(dpy, wwin->client_win, w_global.atom.wmaker.state,
+			w_global.atom.wmaker.state, 32, PropModeReplace, (unsigned char *)data, 10);
 }
 
 static int getSavedState(Window window, WSavedState ** state)
@@ -2312,12 +2401,17 @@ static int getSavedState(Window window, WSavedState ** state)
 	unsigned long bytes_after_ret;
 	long *data;
 
-	if (XGetWindowProperty(dpy, window, _XA_WINDOWMAKER_STATE, 0, 10,
-			       True, _XA_WINDOWMAKER_STATE,
+	if (XGetWindowProperty(dpy, window, w_global.atom.wmaker.state, 0, 10,
+			       True, w_global.atom.wmaker.state,
 			       &type_ret, &fmt_ret, &nitems_ret, &bytes_after_ret,
 			       (unsigned char **)&data) != Success || !data || nitems_ret < 10)
 		return 0;
 
+	if (type_ret != w_global.atom.wmaker.state) {
+		XFree(data);
+		return 0;
+	}
+
 	*state = wmalloc(sizeof(WSavedState));
 
 	(*state)->workspace = data[0];
@@ -2333,13 +2427,10 @@ static int getSavedState(Window window, WSavedState ** state)
 
 	XFree(data);
 
-	if (*state && type_ret == _XA_WINDOWMAKER_STATE)
-		return 1;
-	else
-		return 0;
+	return 1;
 }
 
-#ifdef SHAPE
+#ifdef USE_XSHAPE
 void wWindowClearShape(WWindow * wwin)
 {
 	XShapeCombineMask(dpy, wwin->frame->core->window, ShapeBounding,
@@ -2423,7 +2514,7 @@ void wWindowSetShape(WWindow * wwin)
 			   ShapeBounding, (count > 0 ? ShapeUnion : ShapeSet));
 	XFlush(dpy);
 }
-#endif				/* SHAPE */
+#endif				/* USE_XSHAPE */
 
 /* ====================================================================== */
 
@@ -2537,7 +2628,8 @@ void wWindowUpdateGNUstepAttr(WWindow * wwin, GNUstepWMAttributes * attr)
 	}
 }
 
-WMagicNumber wWindowAddSavedState(char *instance, char *class, char *command, pid_t pid, WSavedState * state)
+WMagicNumber wWindowAddSavedState(const char *instance, const char *class,
+											 const char *command, pid_t pid, WSavedState * state)
 {
 	WWindowState *wstate;
 
@@ -2561,7 +2653,19 @@ WMagicNumber wWindowAddSavedState(char *instance, char *class, char *command, pi
 	return wstate;
 }
 
-#define SAME(x, y) (((x) && (y) && !strcmp((x), (y))) || (!(x) && !(y)))
+static inline int is_same(const char *x, const char *y)
+{
+	if ((x == NULL) && (y == NULL))
+		return 1;
+
+	if ((x == NULL) || (y == NULL))
+		return 0;
+
+	if (strcmp(x, y) == 0)
+		return 1;
+	else
+		return 0;
+}
 
 WMagicNumber wWindowGetSavedState(Window win)
 {
@@ -2577,8 +2681,9 @@ WMagicNumber wWindowGetSavedState(Window win)
 
 	if (PropGetWMClass(win, &class, &instance)) {
 		while (wstate) {
-			if (SAME(instance, wstate->instance) &&
-			    SAME(class, wstate->class) && SAME(command, wstate->command)) {
+			if (is_same(instance, wstate->instance) &&
+			    is_same(class, wstate->class) &&
+			    is_same(command, wstate->command)) {
 				break;
 			}
 			wstate = wstate->next;
@@ -2607,26 +2712,12 @@ void wWindowDeleteSavedState(WMagicNumber id)
 	tmp = windowState;
 	if (tmp == wstate) {
 		windowState = wstate->next;
-		if (wstate->instance)
-			wfree(wstate->instance);
-		if (wstate->class)
-			wfree(wstate->class);
-		if (wstate->command)
-			wfree(wstate->command);
-		wfree(wstate->state);
-		wfree(wstate);
+		release_wwindowstate(wstate);
 	} else {
 		while (tmp->next) {
 			if (tmp->next == wstate) {
 				tmp->next = wstate->next;
-				if (wstate->instance)
-					wfree(wstate->instance);
-				if (wstate->class)
-					wfree(wstate->class);
-				if (wstate->command)
-					wfree(wstate->command);
-				wfree(wstate->state);
-				wfree(wstate);
+				release_wwindowstate(wstate);
 				break;
 			}
 			tmp = tmp->next;
@@ -2645,27 +2736,14 @@ void wWindowDeleteSavedStatesForPID(pid_t pid)
 	if (tmp->pid == pid) {
 		wstate = windowState;
 		windowState = tmp->next;
-		if (wstate->instance)
-			wfree(wstate->instance);
-		if (wstate->class)
-			wfree(wstate->class);
-		if (wstate->command)
-			wfree(wstate->command);
-		wfree(wstate->state);
-		wfree(wstate);
+
+		release_wwindowstate(wstate);
 	} else {
 		while (tmp->next) {
 			if (tmp->next->pid == pid) {
 				wstate = tmp->next;
 				tmp->next = wstate->next;
-				if (wstate->instance)
-					wfree(wstate->instance);
-				if (wstate->class)
-					wfree(wstate->class);
-				if (wstate->command)
-					wfree(wstate->command);
-				wfree(wstate->state);
-				wfree(wstate);
+				release_wwindowstate(wstate);
 				break;
 			}
 			tmp = tmp->next;
@@ -2673,6 +2751,21 @@ void wWindowDeleteSavedStatesForPID(pid_t pid)
 	}
 }
 
+static void release_wwindowstate(WWindowState *wstate)
+{
+	if (wstate->instance)
+		wfree(wstate->instance);
+
+	if (wstate->class)
+		wfree(wstate->class);
+
+	if (wstate->command)
+		wfree(wstate->command);
+
+	wfree(wstate->state);
+	wfree(wstate);
+}
+
 void wWindowSetOmnipresent(WWindow *wwin, Bool flag)
 {
 	if (wwin->flags.omnipresent == flag)
@@ -2686,6 +2779,9 @@ static void resizebarMouseDown(WCoreWindow *sender, void *data, XEvent *event)
 {
 	WWindow *wwin = data;
 
+	/* Parameter not used, but tell the compiler that it is ok */
+	(void) sender;
+
 #ifndef NUMLOCK_HACK
 	if ((event->xbutton.state & ValidModMask)
 	    != (event->xbutton.state & ~LockMask)) {
@@ -2694,7 +2790,7 @@ static void resizebarMouseDown(WCoreWindow *sender, void *data, XEvent *event)
 	}
 #endif
 
-	event->xbutton.state &= ValidModMask;
+	event->xbutton.state &= w_global.shortcut.modifiers_mask;
 
 	CloseWindowMenu(wwin->screen_ptr);
 
@@ -2728,7 +2824,10 @@ static void titlebarDblClick(WCoreWindow *sender, void *data, XEvent *event)
 {
 	WWindow *wwin = data;
 
-	event->xbutton.state &= ValidModMask;
+	/* Parameter not used, but tell the compiler that it is ok */
+	(void) sender;
+
+	event->xbutton.state &= w_global.shortcut.modifiers_mask;
 
 	if (event->xbutton.button == Button1) {
 		if (event->xbutton.state == 0) {
@@ -2766,9 +2865,9 @@ static void titlebarDblClick(WCoreWindow *sender, void *data, XEvent *event)
 			wHideOtherApplications(wwin);
 	} else if (event->xbutton.button == Button2) {
 		wSelectWindow(wwin, !wwin->flags.selected);
-	} else if (event->xbutton.button == WINGsConfiguration.mouseWheelUp) {
+	} else if (event->xbutton.button == W_getconf_mouseWheelUp()) {
 		wShadeWindow(wwin);
-	} else if (event->xbutton.button == WINGsConfiguration.mouseWheelDown) {
+	} else if (event->xbutton.button == W_getconf_mouseWheelDown()) {
 		wUnshadeWindow(wwin);
 	}
 }
@@ -2792,7 +2891,7 @@ static void frameMouseDown(WObjDescriptor *desc, XEvent *event)
 		resize_height_increment = wPreferences.resize_increment;
 	}
 
-	event->xbutton.state &= ValidModMask;
+	event->xbutton.state &= w_global.shortcut.modifiers_mask;
 
 	CloseWindowMenu(wwin->screen_ptr);
 
@@ -2843,12 +2942,15 @@ static void titlebarMouseDown(WCoreWindow *sender, void *data, XEvent *event)
 {
 	WWindow *wwin = (WWindow *) data;
 
+	/* Parameter not used, but tell the compiler that it is ok */
+	(void) sender;
+
 #ifndef NUMLOCK_HACK
 	if ((event->xbutton.state & ValidModMask) != (event->xbutton.state & ~LockMask))
 		wwarning(_("The NumLock, ScrollLock or similar key seems to be turned on. "
 			   "Turn it off or some mouse actions and keyboard shortcuts will not work."));
 #endif
-	event->xbutton.state &= ValidModMask;
+	event->xbutton.state &= w_global.shortcut.modifiers_mask;
 
 	CloseWindowMenu(wwin->screen_ptr);
 
@@ -2906,7 +3008,10 @@ static void windowCloseClick(WCoreWindow *sender, void *data, XEvent *event)
 {
 	WWindow *wwin = data;
 
-	event->xbutton.state &= ValidModMask;
+	/* Parameter not used, but tell the compiler that it is ok */
+	(void) sender;
+
+	event->xbutton.state &= w_global.shortcut.modifiers_mask;
 
 	CloseWindowMenu(wwin->screen_ptr);
 
@@ -2919,7 +3024,8 @@ static void windowCloseClick(WCoreWindow *sender, void *data, XEvent *event)
 	} else {
 		if (wwin->protocols.DELETE_WINDOW && event->xbutton.state == 0) {
 			/* send delete message */
-			wClientSendProtocol(wwin, _XA_WM_DELETE_WINDOW, LastTimestamp);
+			wClientSendProtocol(wwin, w_global.atom.wm.delete_window,
+									  w_global.timestamp.last_event);
 		}
 	}
 }
@@ -2928,6 +3034,9 @@ static void windowCloseDblClick(WCoreWindow *sender, void *data, XEvent *event)
 {
 	WWindow *wwin = data;
 
+	/* Parameter not used, but tell the compiler that it is ok */
+	(void) sender;
+
 	CloseWindowMenu(wwin->screen_ptr);
 
 	if (event->xbutton.button < Button1 || event->xbutton.button > Button3)
@@ -2935,7 +3044,8 @@ static void windowCloseDblClick(WCoreWindow *sender, void *data, XEvent *event)
 
 	/* send delete message */
 	if (wwin->protocols.DELETE_WINDOW)
-		wClientSendProtocol(wwin, _XA_WM_DELETE_WINDOW, LastTimestamp);
+		wClientSendProtocol(wwin, w_global.atom.wm.delete_window,
+								  w_global.timestamp.last_event);
 	else
 		wClientKill(wwin);
 }
@@ -2948,6 +3058,9 @@ static void windowLanguageClick(WCoreWindow *sender, void *data, XEvent *event)
 	WScreen *scr = fwin->screen_ptr;
 	int tl;
 
+	/* Parameter not used, but tell the compiler that it is ok */
+	(void) sender;
+
 	if (event->xbutton.button != Button1 && event->xbutton.button != Button3)
 		return;
 	tl = wwin->frame->languagemode;
@@ -2967,7 +3080,10 @@ static void windowIconifyClick(WCoreWindow *sender, void *data, XEvent *event)
 {
 	WWindow *wwin = data;
 
-	event->xbutton.state &= ValidModMask;
+	/* Parameter not used, but tell the compiler that it is ok */
+	(void) sender;
+
+	event->xbutton.state &= w_global.shortcut.modifiers_mask;
 
 	CloseWindowMenu(wwin->screen_ptr);
 
@@ -2975,7 +3091,8 @@ static void windowIconifyClick(WCoreWindow *sender, void *data, XEvent *event)
 		return;
 
 	if (wwin->protocols.MINIATURIZE_WINDOW && event->xbutton.state == 0) {
-		wClientSendProtocol(wwin, _XA_GNUSTEP_WM_MINIATURIZE_WINDOW, LastTimestamp);
+		wClientSendProtocol(wwin, w_global.atom.gnustep.wm_miniaturize_window,
+								  w_global.timestamp.last_event);
 	} else {
 		WApplication *wapp;
 		if ((event->xbutton.state & ControlMask) || (event->xbutton.button == Button3)) {
