@@ -28,15 +28,17 @@
 
 #include <X11/Xlib.h>
 #include <X11/Xutil.h>
+#include <X11/XKBlib.h>
 
 #include "WindowMaker.h"
 #include "actions.h"
 #include "menu.h"
-#include "funcs.h"
+#include "main.h"
 #include "window.h"
 #include "client.h"
 #include "application.h"
 #include "keybind.h"
+#include "misc.h"
 #include "framewin.h"
 #include "workspace.h"
 #include "winspector.h"
@@ -44,35 +46,95 @@
 #include "stacking.h"
 #include "icon.h"
 #include "xinerama.h"
+#include "winmenu.h"
 
-#define MC_MAXIMIZE	0
-#define MC_MINIATURIZE	1
-#define MC_SHADE	2
-#define MC_HIDE		3
-#define MC_MOVERESIZE   4
-#define MC_SELECT       5
-#define MC_DUMMY_MOVETO 6
-#define MC_PROPERTIES   7
-#define MC_OPTIONS      8
-#define MC_SHORTCUT     8
-#define MC_RELAUNCH     9
 
-#define MC_CLOSE        10
-#define MC_KILL         11
+static WMenu *makeWorkspaceMenu(WScreen *scr);
+static WMenu *makeOptionsMenu(WScreen *scr);
+static WMenu *makeMaximizeMenu(WScreen *scr);
 
-#define WO_KEEP_ON_TOP		0
-#define WO_KEEP_AT_BOTTOM	1
-#define WO_OMNIPRESENT		2
-#define WO_ENTRIES		3
+/*
+ * Define the Menu entry that will be present in the Window menu
+ *
+ * The order of this list defines the order in which they will appear;
+ * make sure to keep the structure below aligned with the list because
+ * the constant index is used in many places.
+ */
+enum
+{
+	MC_MAXIMIZE,
+	MC_OTHERMAX,
+	MC_MINIATURIZE,
+	MC_SHADE,
+	MC_HIDE,
+	MC_MOVERESIZE,
+	MC_SELECT,
+	MC_CHANGEWKSPC,
+	MC_PROPERTIES,
+	MC_OPTIONS,
+	MC_RELAUNCH,
+	MC_CLOSE,
+	MC_KILL
+};
 
-/**** Global data ***/
-extern Time LastTimestamp;
-extern Atom _XA_WM_DELETE_WINDOW;
-extern Atom _XA_GNUSTEP_WM_MINIATURIZE_WINDOW;
+static const struct {
+	const char *label;
+	WMenu *(*generate_submenu)(WScreen *scr);
+} window_menu_entries[] = {
+	[MC_MAXIMIZE]    = { N_("Maximize"), NULL },
+	[MC_OTHERMAX]    = { N_("Other maximization"), makeMaximizeMenu },
+	[MC_MINIATURIZE] = { N_("Miniaturize"), NULL },
+	[MC_SHADE]       = { N_("Shade"), NULL },
+	[MC_HIDE]        = { N_("Hide"), NULL },
+	[MC_MOVERESIZE]  = { N_("Resize/Move"), NULL },
+	[MC_SELECT]      = { N_("Select"), NULL },
+	[MC_CHANGEWKSPC] = { N_("Move To"), makeWorkspaceMenu },
+	[MC_PROPERTIES]  = { N_("Attributes..."), NULL },
+	[MC_OPTIONS]     = { N_("Options"), makeOptionsMenu },
+	[MC_RELAUNCH]    = { N_("Launch"), NULL },
+	[MC_CLOSE]       = { N_("Close"), NULL },
+	[MC_KILL]        = { N_("Kill"), NULL }
+};
 
-extern WShortKey wKeyBindings[WKBD_LAST];
+/*
+ * Defines the menu entries for the Options sub-menu
+ *
+ * These options will be placed at the beginning of the menu, the rest will
+ * be populated with the Window Shortcut possibilities
+ */
+enum
+{
+	WO_KEEP_ON_TOP,
+	WO_KEEP_AT_BOTTOM,
+	WO_OMNIPRESENT
+};
 
-extern WPreferences wPreferences;
+static const char *const menu_options_entries[] = {
+	[WO_KEEP_ON_TOP]    = N_("Keep on top"),
+	[WO_KEEP_AT_BOTTOM] = N_("Keep at bottom"),
+	[WO_OMNIPRESENT]    = N_("Omnipresent")
+};
+
+/*
+ * Defines the menu entries for the Other maximization sub-menu
+ */
+static const struct {
+	const char *label;
+	unsigned int shortcut_idx;
+	int maxim_direction;
+} menu_maximize_entries[] = {
+	{ N_("Maximize vertically"), WKBD_VMAXIMIZE, MAX_VERTICAL },
+	{ N_("Maximize horizontally"), WKBD_HMAXIMIZE, MAX_HORIZONTAL },
+	{ N_("Maximize left half"), WKBD_LHMAXIMIZE, MAX_VERTICAL | MAX_LEFTHALF },
+	{ N_("Maximize right half"), WKBD_RHMAXIMIZE, MAX_VERTICAL | MAX_RIGHTHALF },
+	{ N_("Maximize top half"), WKBD_THMAXIMIZE, MAX_HORIZONTAL | MAX_TOPHALF },
+	{ N_("Maximize bottom half"), WKBD_BHMAXIMIZE, MAX_HORIZONTAL | MAX_BOTTOMHALF },
+	{ N_("Maximize left top corner"), WKBD_LTCMAXIMIZE, MAX_LEFTHALF | MAX_TOPHALF },
+	{ N_("Maximize right top corner"), WKBD_RTCMAXIMIZE, MAX_RIGHTHALF | MAX_TOPHALF },
+	{ N_("Maximize left bottom corner"), WKBD_LBCMAXIMIZE, MAX_LEFTHALF | MAX_BOTTOMHALF },
+	{ N_("Maximize right bottom corner"), WKBD_RBCMAXIMIZE, MAX_RIGHTHALF | MAX_BOTTOMHALF },
+	{ N_("Maximus: tiled maximization"), WKBD_MAXIMUS, MAX_MAXIMUS }
+};
 
 static void updateOptionsMenu(WMenu * menu, WWindow * wwin);
 
@@ -80,6 +142,9 @@ static void execWindowOptionCommand(WMenu * menu, WMenuEntry * entry)
 {
 	WWindow *wwin = (WWindow *) entry->clientdata;
 
+	/* Parameter not used, but tell the compiler that it is ok */
+	(void) menu;
+
 	switch (entry->order) {
 	case WO_KEEP_ON_TOP:
 		if (wwin->frame->core->stacking->window_level != WMFloatingLevel)
@@ -101,6 +166,73 @@ static void execWindowOptionCommand(WMenu * menu, WMenuEntry * entry)
 	}
 }
 
+static void execMaximizeCommand(WMenu * menu, WMenuEntry * entry)
+{
+	WWindow *wwin = (WWindow *) entry->clientdata;
+	
+	/* Parameter not used, but tell the compiler that it is ok */
+	(void) menu;
+
+	handleMaximize(wwin, menu_maximize_entries[entry->order].maxim_direction);
+}
+
+static void updateUnmaximizeShortcut(WMenuEntry * entry, int flags)
+{
+	int key;
+
+	switch (flags & (MAX_HORIZONTAL | MAX_VERTICAL | MAX_LEFTHALF | MAX_RIGHTHALF | MAX_TOPHALF | MAX_BOTTOMHALF | MAX_MAXIMUS)) {
+	case MAX_HORIZONTAL:
+		key = WKBD_HMAXIMIZE;
+		break;
+
+	case MAX_VERTICAL:
+		key = WKBD_VMAXIMIZE;
+		break;
+
+	case MAX_LEFTHALF | MAX_VERTICAL:
+		key = WKBD_LHMAXIMIZE;
+		break;
+
+	case MAX_RIGHTHALF | MAX_VERTICAL:
+		key = WKBD_RHMAXIMIZE;
+		break;
+
+	case MAX_TOPHALF | MAX_HORIZONTAL:
+		key = WKBD_THMAXIMIZE;
+		break;
+
+	case MAX_BOTTOMHALF | MAX_HORIZONTAL:
+		key = WKBD_BHMAXIMIZE;
+		break;
+
+	case MAX_LEFTHALF | MAX_TOPHALF:
+		key = WKBD_LTCMAXIMIZE;
+		break;
+
+	case MAX_RIGHTHALF | MAX_TOPHALF:
+		key = WKBD_RTCMAXIMIZE;
+		break;
+
+	case MAX_LEFTHALF | MAX_BOTTOMHALF:
+		key = WKBD_LBCMAXIMIZE;
+		break;
+
+	case MAX_RIGHTHALF | MAX_BOTTOMHALF:
+		key = WKBD_RBCMAXIMIZE;
+		break;
+
+	case MAX_MAXIMUS:
+		key = WKBD_MAXIMUS;
+		break;
+
+	default:
+		key = WKBD_MAXIMIZE;
+		break;
+	}
+
+	entry->rtext = GetShortcutKey(wKeyBindings[key]);
+}
+
 static void execMenuCommand(WMenu * menu, WMenuEntry * entry)
 {
 	WWindow *wwin = (WWindow *) entry->clientdata;
@@ -111,7 +243,8 @@ static void execMenuCommand(WMenu * menu, WMenuEntry * entry)
 	switch (entry->order) {
 	case MC_CLOSE:
 		/* send delete message */
-		wClientSendProtocol(wwin, _XA_WM_DELETE_WINDOW, LastTimestamp);
+		wClientSendProtocol(wwin, w_global.atom.wm.delete_window,
+								  w_global.timestamp.last_event);
 		break;
 
 	case MC_KILL:
@@ -132,7 +265,8 @@ static void execMenuCommand(WMenu * menu, WMenuEntry * entry)
 			wDeiconifyWindow(wwin);
 		} else {
 			if (wwin->protocols.MINIATURIZE_WINDOW) {
-				wClientSendProtocol(wwin, _XA_GNUSTEP_WM_MINIATURIZE_WINDOW, LastTimestamp);
+				wClientSendProtocol(wwin, w_global.atom.gnustep.wm_miniaturize_window,
+										  w_global.timestamp.last_event);
 			} else {
 				wIconifyWindow(wwin);
 			}
@@ -183,6 +317,9 @@ static void switchWSCommand(WMenu * menu, WMenuEntry * entry)
 {
 	WWindow *wwin = (WWindow *) entry->clientdata;
 
+	/* Parameter not used, but tell the compiler that it is ok */
+	(void) menu;
+
 	wSelectWindow(wwin, False);
 	wWindowChangeWorkspace(wwin, entry->order);
 }
@@ -191,7 +328,10 @@ static void makeShortcutCommand(WMenu * menu, WMenuEntry * entry)
 {
 	WWindow *wwin = (WWindow *) entry->clientdata;
 	WScreen *scr = wwin->screen_ptr;
-	int index = entry->order - WO_ENTRIES;
+	int index = entry->order - wlengthof(menu_options_entries);
+
+	/* Parameter not used, but tell the compiler that it is ok */
+	(void) menu;
 
 	if (scr->shortcutWindows[index]) {
 		WMFreeArray(scr->shortcutWindows[index]);
@@ -218,53 +358,45 @@ static void updateWorkspaceMenu(WMenu * menu)
 {
 	WScreen *scr = menu->frame->screen_ptr;
 	char title[MAX_WORKSPACENAME_WIDTH + 1];
+	WMenuEntry *entry;
 	int i;
 
 	for (i = 0; i < scr->workspace_count; i++) {
 		if (i < menu->entry_no) {
-			if (strcmp(menu->entries[i]->text, scr->workspaces[i]->name) != 0) {
-				wfree(menu->entries[i]->text);
+
+			entry = menu->entries[i];
+			if (strcmp(entry->text, scr->workspaces[i]->name) != 0) {
+				wfree(entry->text);
 				strncpy(title, scr->workspaces[i]->name, MAX_WORKSPACENAME_WIDTH);
 				title[MAX_WORKSPACENAME_WIDTH] = 0;
 				menu->entries[i]->text = wstrdup(title);
+				menu->entries[i]->rtext = GetShortcutKey(wKeyBindings[WKBD_MOVE_WORKSPACE1 + i]);
 				menu->flags.realized = 0;
 			}
 		} else {
 			strncpy(title, scr->workspaces[i]->name, MAX_WORKSPACENAME_WIDTH);
 			title[MAX_WORKSPACENAME_WIDTH] = 0;
 
-			wMenuAddCallback(menu, title, switchWSCommand, NULL);
+			entry = wMenuAddCallback(menu, title, switchWSCommand, NULL);
+			entry->rtext = GetShortcutKey(wKeyBindings[WKBD_MOVE_WORKSPACE1 + i]);
 
 			menu->flags.realized = 0;
 		}
+
+		/* workspace shortcut labels */
+		if (i / 10 == scr->current_workspace / 10)
+			entry->rtext = GetShortcutKey(wKeyBindings[WKBD_MOVE_WORKSPACE1 + (i % 10)]);
+		else
+			entry->rtext = NULL;
 	}
 
 	if (!menu->flags.realized)
 		wMenuRealize(menu);
 }
 
-static char *getShortcutString(WShortKey key)
-{
-	char *tmp = NULL;
-	char *k = XKeysymToString(XKeycodeToKeysym(dpy, key.keycode, 0));
-	if (!k) return NULL;
-
-	char **m = wPreferences.modifier_labels;
-	if (key.modifier & ControlMask) tmp = wstrappend(tmp, m[1] ? m[1] : "Ctrl+");
-	if (key.modifier & ShiftMask)   tmp = wstrappend(tmp, m[0] ? m[0] : "Shift+");
-	if (key.modifier & Mod1Mask)    tmp = wstrappend(tmp, m[2] ? m[2] : "Mod1+");
-	if (key.modifier & Mod2Mask)    tmp = wstrappend(tmp, m[3] ? m[3] : "Mod2+");
-	if (key.modifier & Mod3Mask)    tmp = wstrappend(tmp, m[4] ? m[4] : "Mod3+");
-	if (key.modifier & Mod4Mask)    tmp = wstrappend(tmp, m[5] ? m[5] : "Mod4+");
-	if (key.modifier & Mod5Mask)    tmp = wstrappend(tmp, m[6] ? m[6] : "Mod5+");
-	tmp = wstrappend(tmp, k);
-
-	return tmp;
-}
-
 static void updateMakeShortcutMenu(WMenu * menu, WWindow * wwin)
 {
-	WMenu *smenu = menu->cascades[menu->entries[MC_SHORTCUT]->cascade];
+	WMenu *smenu = menu->cascades[menu->entries[MC_OPTIONS]->cascade];
 	int i;
 	char *buffer;
 	int buflen;
@@ -276,9 +408,8 @@ static void updateMakeShortcutMenu(WMenu * menu, WWindow * wwin)
 	buflen = strlen(_("Set Shortcut")) + 16;
 	buffer = wmalloc(buflen);
 
-	for (i = WO_ENTRIES; i < smenu->entry_no; i++) {
-		char *tmp;
-		int shortcutNo = i - WO_ENTRIES;
+	for (i = wlengthof(menu_options_entries); i < smenu->entry_no; i++) {
+		int shortcutNo = i - wlengthof(menu_options_entries);
 		WMenuEntry *entry = smenu->entries[i];
 		WMArray *shortSelWindows = wwin->screen_ptr->shortcutWindows[shortcutNo];
 
@@ -303,12 +434,28 @@ static void updateMakeShortcutMenu(WMenu * menu, WWindow * wwin)
 		kcode = wKeyBindings[WKBD_WINDOW1 + shortcutNo].keycode;
 
 		if (kcode) {
-			if ((tmp = getShortcutString(wKeyBindings[WKBD_WINDOW1 + shortcutNo]))
-			    && (!entry->rtext || strcmp(tmp, entry->rtext) != 0)) {
-				if (entry->rtext)
+			char *tmp;
+
+			tmp = GetShortcutKey(wKeyBindings[WKBD_WINDOW1 + shortcutNo]);
+			if (tmp == NULL) {
+				if (entry->rtext != NULL) {
+					/* There was a shortcut, but there is no more */
 					wfree(entry->rtext);
+					entry->rtext = NULL;
+					smenu->flags.realized = 0;
+				}
+			} else if (entry->rtext == NULL) {
+				/* There was no shortcut, but there is one now */
 				entry->rtext = tmp;
 				smenu->flags.realized = 0;
+			} else if (strcmp(tmp, entry->rtext) != 0) {
+				/* There was a shortcut, but it has changed */
+				wfree(entry->rtext);
+				entry->rtext = tmp;
+				smenu->flags.realized = 0;
+			} else {
+				/* There was a shortcut but it did not change */
+				wfree(tmp);
 			}
 			wMenuSetEnabled(smenu, i, True);
 		} else {
@@ -350,6 +497,20 @@ static void updateOptionsMenu(WMenu * menu, WWindow * wwin)
 	wMenuRealize(smenu);
 }
 
+static void updateMaximizeMenu(WMenu * menu, WWindow * wwin)
+{
+	WMenu *smenu = menu->cascades[menu->entries[MC_OTHERMAX]->cascade];
+	int i;
+
+	for (i = 0; i < smenu->entry_no; i++) {
+		smenu->entries[i]->clientdata = wwin;
+		smenu->entries[i]->rtext = GetShortcutKey(wKeyBindings[menu_maximize_entries[i].shortcut_idx]);
+	}
+
+	smenu->flags.realized = 0;
+	wMenuRealize(smenu);
+}
+
 static WMenu *makeWorkspaceMenu(WScreen * scr)
 {
 	WMenu *menu;
@@ -362,29 +523,13 @@ static WMenu *makeWorkspaceMenu(WScreen * scr)
 
 	updateWorkspaceMenu(menu);
 
-	return menu;
-}
-
-static WMenu *makeMakeShortcutMenu(WScreen * scr, WMenu * menu)
-{
 	/*
-	   WMenu *menu;
+	 * The Workspace Menu is made visible in the screen structure because
+	 * it is updated when there is a change on workspaces. This was done
+	 * to be efficient, avoiding re-generating completely the window menu
+	 * and its sub-menus everytime it is needed.
 	 */
-	int i;
-	/*
-	   menu = wMenuCreate(scr, NULL, False);
-	   if (!menu) {
-	   wwarning(_("could not create submenu for window menu"));
-	   return NULL;
-	   }
-	 */
-
-	for (i = 0; i < MAX_WINDOW_SHORTCUTS; i++) {
-		WMenuEntry *entry;
-		entry = wMenuAddCallback(menu, "", makeShortcutCommand, NULL);
-
-		entry->flags.indicator = 1;
-	}
+	scr->workspace_submenu = menu;
 
 	return menu;
 }
@@ -393,6 +538,7 @@ static WMenu *makeOptionsMenu(WScreen * scr)
 {
 	WMenu *menu;
 	WMenuEntry *entry;
+	int i;
 
 	menu = wMenuCreate(scr, NULL, False);
 	if (!menu) {
@@ -400,17 +546,33 @@ static WMenu *makeOptionsMenu(WScreen * scr)
 		return NULL;
 	}
 
-	entry = wMenuAddCallback(menu, _("Keep on top"), execWindowOptionCommand, NULL);
-	entry->flags.indicator = 1;
-	entry->flags.indicator_type = MI_CHECK;
+	for (i = 0; i < wlengthof(menu_options_entries); i++) {
+		entry = wMenuAddCallback(menu, _(menu_options_entries[i]), execWindowOptionCommand, NULL);
+		entry->flags.indicator = 1;
+		entry->flags.indicator_type = MI_CHECK;
+	}
 
-	entry = wMenuAddCallback(menu, _("Keep at bottom"), execWindowOptionCommand, NULL);
-	entry->flags.indicator = 1;
-	entry->flags.indicator_type = MI_CHECK;
+	for (i = 0; i < MAX_WINDOW_SHORTCUTS; i++) {
+		entry = wMenuAddCallback(menu, "", makeShortcutCommand, NULL);
+		entry->flags.indicator = 1;
+	}
 
-	entry = wMenuAddCallback(menu, _("Omnipresent"), execWindowOptionCommand, NULL);
-	entry->flags.indicator = 1;
-	entry->flags.indicator_type = MI_CHECK;
+	return menu;
+}
+
+static WMenu *makeMaximizeMenu(WScreen * scr)
+{
+	WMenu *menu;
+	int i;
+
+	menu = wMenuCreate(scr, NULL, False);
+	if (!menu) {
+		wwarning(_("could not create submenu for window menu"));
+		return NULL;
+	}
+
+	for (i = 0; i < wlengthof(menu_maximize_entries); i++)
+		wMenuAddCallback(menu, _(menu_maximize_entries[i].label), execMaximizeCommand, NULL);
 
 	return menu;
 }
@@ -418,54 +580,23 @@ static WMenu *makeOptionsMenu(WScreen * scr)
 static WMenu *createWindowMenu(WScreen * scr)
 {
 	WMenu *menu;
-	WMenuEntry *entry;
+	int i;
 
 	menu = wMenuCreate(scr, NULL, False);
-	/*
-	 * Warning: If you make some change that affects the order of the
-	 * entries, you must update the command #defines in the top of
-	 * this file.
-	 */
-	entry = wMenuAddCallback(menu, _("Maximize"), execMenuCommand, NULL);
-	entry->rtext = getShortcutString(wKeyBindings[WKBD_MAXIMIZE]);
 
-	entry = wMenuAddCallback(menu, _("Miniaturize"), execMenuCommand, NULL);
-	entry->rtext = getShortcutString(wKeyBindings[WKBD_MINIATURIZE]);
+	for (i = 0; i < wlengthof(window_menu_entries); i++) {
+		WMenuEntry *entry;
 
-	entry = wMenuAddCallback(menu, _("Shade"), execMenuCommand, NULL);
-	entry->rtext = getShortcutString(wKeyBindings[WKBD_SHADE]);
+		entry = wMenuAddCallback(menu, _(window_menu_entries[i].label),
+		                         (window_menu_entries[i].generate_submenu == NULL)?execMenuCommand:NULL,
+		                         NULL);
+		if (window_menu_entries[i].generate_submenu != NULL) {
+			WMenu *submenu;
 
-	entry = wMenuAddCallback(menu, _("Hide"), execMenuCommand, NULL);
-	entry->rtext = getShortcutString(wKeyBindings[WKBD_HIDE]);
-
-	entry = wMenuAddCallback(menu, _("Resize/Move"), execMenuCommand, NULL);
-	entry->rtext = getShortcutString(wKeyBindings[WKBD_MOVERESIZE]);
-
-	entry = wMenuAddCallback(menu, _("Select"), execMenuCommand, NULL);
-	entry->rtext = getShortcutString(wKeyBindings[WKBD_SELECT]);
-
-	entry = wMenuAddCallback(menu, _("Move To"), NULL, NULL);
-	scr->workspace_submenu = makeWorkspaceMenu(scr);
-	if (scr->workspace_submenu)
-		wMenuEntrySetCascade(menu, entry, scr->workspace_submenu);
-
-	entry = wMenuAddCallback(menu, _("Attributes..."), execMenuCommand, NULL);
-
-	entry = wMenuAddCallback(menu, _("Options"), NULL, NULL);
-	wMenuEntrySetCascade(menu, entry, makeMakeShortcutMenu(scr, makeOptionsMenu(scr)));
-
-	/*
-	   entry = wMenuAddCallback(menu, _("Select Shortcut"), NULL, NULL);
-	   wMenuEntrySetCascade(menu, entry, makeMakeShortcutMenu(scr));
-	 */
-
-	entry = wMenuAddCallback(menu, _("Launch"), execMenuCommand, NULL);
-	entry->rtext = getShortcutString(wKeyBindings[WKBD_RELAUNCH]);
-
-	entry = wMenuAddCallback(menu, _("Close"), execMenuCommand, NULL);
-	entry->rtext = getShortcutString(wKeyBindings[WKBD_CLOSE]);
-
-	entry = wMenuAddCallback(menu, _("Kill"), execMenuCommand, NULL);
+			submenu = window_menu_entries[i].generate_submenu(scr);
+			wMenuEntrySetCascade(menu, entry, submenu);
+		}
+	}
 
 	return menu;
 }
@@ -492,6 +623,7 @@ static void updateMenuForWindow(WMenu * menu, WWindow * wwin)
 	int i;
 
 	updateOptionsMenu(menu, wwin);
+	updateMaximizeMenu(menu, wwin);
 
 	updateMakeShortcutMenu(menu, wwin);
 
@@ -521,12 +653,14 @@ static void updateMenuForWindow(WMenu * menu, WWindow * wwin)
 			text = _("Unmaximize");
 
 		menu->entries[MC_MAXIMIZE]->text = text;
+		updateUnmaximizeShortcut(menu->entries[MC_MAXIMIZE], wwin->flags.maximized);
 	} else {
 		static char *text = NULL;
 		if (!text)
 			text = _("Maximize");
 
 		menu->entries[MC_MAXIMIZE]->text = text;
+		menu->entries[MC_MAXIMIZE]->rtext = GetShortcutKey(wKeyBindings[WKBD_MAXIMIZE]);
 	}
 	wMenuSetEnabled(menu, MC_MAXIMIZE, IS_RESIZABLE(wwin));
 
@@ -550,7 +684,21 @@ static void updateMenuForWindow(WMenu * menu, WWindow * wwin)
 	wMenuSetEnabled(menu, MC_SHADE, !WFLAGP(wwin, no_shadeable)
 			&& !wwin->flags.miniaturized);
 
-	wMenuSetEnabled(menu, MC_DUMMY_MOVETO, !IS_OMNIPRESENT(wwin));
+	if (wwin->flags.selected) {
+		static char *text = NULL;
+		if (!text)
+			text = _("Deselect");
+
+		menu->entries[MC_SELECT]->text = text;
+	} else {
+		static char *text = NULL;
+		if (!text)
+			text = _("Select");
+
+		menu->entries[MC_SELECT]->text = text;
+	}
+
+	wMenuSetEnabled(menu, MC_CHANGEWKSPC, !IS_OMNIPRESENT(wwin));
 
 	if (!wwin->flags.inspector_open) {
 		wMenuSetEnabled(menu, MC_PROPERTIES, True);
@@ -558,6 +706,17 @@ static void updateMenuForWindow(WMenu * menu, WWindow * wwin)
 		wMenuSetEnabled(menu, MC_PROPERTIES, False);
 	}
 
+	/* Update shortcut labels except for (Un)Maximize which is
+	 * handled separately.
+	 */
+	menu->entries[MC_MINIATURIZE]->rtext = GetShortcutKey(wKeyBindings[WKBD_MINIATURIZE]);
+	menu->entries[MC_SHADE]->rtext = GetShortcutKey(wKeyBindings[WKBD_SHADE]);
+	menu->entries[MC_HIDE]->rtext = GetShortcutKey(wKeyBindings[WKBD_HIDE]);
+	menu->entries[MC_MOVERESIZE]->rtext = GetShortcutKey(wKeyBindings[WKBD_MOVERESIZE]);
+	menu->entries[MC_SELECT]->rtext = GetShortcutKey(wKeyBindings[WKBD_SELECT]);
+	menu->entries[MC_RELAUNCH]->rtext = GetShortcutKey(wKeyBindings[WKBD_RELAUNCH]);
+	menu->entries[MC_CLOSE]->rtext = GetShortcutKey(wKeyBindings[WKBD_CLOSE]);
+
 	/* set the client data of the entries to the window */
 	for (i = 0; i < menu->entry_no; i++) {
 		menu->entries[i]->clientdata = wwin;
@@ -565,22 +724,20 @@ static void updateMenuForWindow(WMenu * menu, WWindow * wwin)
 
 	for (i = 0; i < scr->workspace_submenu->entry_no; i++) {
 		scr->workspace_submenu->entries[i]->clientdata = wwin;
-		if (i == scr->current_workspace) {
+		if (i == scr->current_workspace)
 			wMenuSetEnabled(scr->workspace_submenu, i, False);
-		} else {
+		else
 			wMenuSetEnabled(scr->workspace_submenu, i, True);
-		}
 	}
 
 	menu->flags.realized = 0;
 	wMenuRealize(menu);
 }
 
-void OpenWindowMenu(WWindow * wwin, int x, int y, int keyboard)
+static WMenu *open_window_menu_core(WWindow *wwin)
 {
-	WMenu *menu;
 	WScreen *scr = wwin->screen_ptr;
-	WMRect rect;
+	WMenu *menu;
 
 	wwin->flags.menu_open_for_me = 1;
 
@@ -591,6 +748,7 @@ void OpenWindowMenu(WWindow * wwin, int x, int y, int keyboard)
 		wfree(scr->window_menu->entries[MC_MINIATURIZE]->text);
 		wfree(scr->window_menu->entries[MC_MAXIMIZE]->text);
 		wfree(scr->window_menu->entries[MC_SHADE]->text);
+		wfree(scr->window_menu->entries[MC_SELECT]->text);
 	} else {
 		updateWorkspaceMenu(scr->workspace_submenu);
 	}
@@ -598,25 +756,44 @@ void OpenWindowMenu(WWindow * wwin, int x, int y, int keyboard)
 	menu = scr->window_menu;
 	if (menu->flags.mapped) {
 		wMenuUnmap(menu);
-		if (menu->entries[0]->clientdata == wwin) {
-			return;
-		}
+		if (menu->entries[0]->clientdata == wwin)
+			return NULL;
 	}
 
 	updateMenuForWindow(menu, wwin);
 
+	return menu;
+}
+
+static void prepare_menu_position(WMenu *menu, int *x, int *y)
+{
+	WMRect rect;
+
+	rect = wGetRectForHead(menu->frame->screen_ptr,
+			       wGetHeadForPointerLocation(menu->frame->screen_ptr));
+	if (*x < rect.pos.x - menu->frame->core->width / 2)
+		*x = rect.pos.x - menu->frame->core->width / 2;
+	if (*y < rect.pos.y)
+		*y = rect.pos.y;
+}
+
+void OpenWindowMenu(WWindow *wwin, int x, int y, int keyboard)
+{
+	WMenu *menu;
+
+	menu = open_window_menu_core(wwin);
+	if (!menu)
+		return;
+
+	/* Specific menu position */
 	x -= menu->frame->core->width / 2;
 	if (x + menu->frame->core->width > wwin->frame_x + wwin->frame->core->width)
 		x = wwin->frame_x + wwin->frame->core->width - menu->frame->core->width;
 	if (x < wwin->frame_x)
 		x = wwin->frame_x;
 
-	rect = wGetRectForHead(menu->frame->screen_ptr,
-			       wGetHeadForPointerLocation(menu->frame->screen_ptr));
-	if (x < rect.pos.x - menu->frame->core->width / 2)
-		x = rect.pos.x - menu->frame->core->width / 2;
-	if (y < rect.pos.y)
-		y = rect.pos.y;
+	/* Common menu position */
+	prepare_menu_position(menu, &x, &y);
 
 	if (!wwin->flags.internal_window)
 		wMenuMapAt(menu, x, y, keyboard);
@@ -627,31 +804,12 @@ void OpenWindowMenu2(WWindow *wwin, int x, int y, int keyboard)
 	int i;
 	WMenu *menu;
 	WScreen *scr = wwin->screen_ptr;
-	WMRect rect;
 
-	wwin->flags.menu_open_for_me = 1;
-
-	if (!scr->window_menu) {
-		scr->window_menu = createWindowMenu(scr);
-
-		/* hack to save some memory allocation/deallocation */
-		wfree(scr->window_menu->entries[MC_MINIATURIZE]->text);
-		wfree(scr->window_menu->entries[MC_MAXIMIZE]->text);
-		wfree(scr->window_menu->entries[MC_SHADE]->text);
-	} else {
-		updateWorkspaceMenu(scr->workspace_submenu);
-	}
-
-	menu = scr->window_menu;
-	if (menu->flags.mapped) {
-		wMenuUnmap(menu);
-		if (menu->entries[0]->clientdata == wwin) {
-			return;
-		}
-	}
-
-	updateMenuForWindow(menu, wwin);
+	menu = open_window_menu_core(wwin);
+	if (!menu)
+		return;
 
+	/* Specific menu position */
 	for (i = 0; i < scr->workspace_submenu->entry_no; i++) {
 		scr->workspace_submenu->entries[i]->clientdata = wwin;
 		wMenuSetEnabled(scr->workspace_submenu, i, True);
@@ -659,12 +817,8 @@ void OpenWindowMenu2(WWindow *wwin, int x, int y, int keyboard)
 
 	x -= menu->frame->core->width / 2;
 
-	rect = wGetRectForHead(menu->frame->screen_ptr,
-			       wGetHeadForPointerLocation(menu->frame->screen_ptr));
-	if (x < rect.pos.x - menu->frame->core->width / 2)
-		x = rect.pos.x - menu->frame->core->width / 2;
-	if (y < rect.pos.y)
-		y = rect.pos.y;
+	/* Common menu position */
+	prepare_menu_position(menu, &x, &y);
 
 	if (!wwin->flags.internal_window)
 		wMenuMapAt(menu, x, y, keyboard);
@@ -673,30 +827,10 @@ void OpenWindowMenu2(WWindow *wwin, int x, int y, int keyboard)
 void OpenMiniwindowMenu(WWindow * wwin, int x, int y)
 {
 	WMenu *menu;
-	WScreen *scr = wwin->screen_ptr;
 
-	wwin->flags.menu_open_for_me = 1;
-
-	if (!scr->window_menu) {
-		scr->window_menu = createWindowMenu(scr);
-
-		/* hack to save some memory allocation/deallocation */
-		wfree(scr->window_menu->entries[MC_MINIATURIZE]->text);
-		wfree(scr->window_menu->entries[MC_MAXIMIZE]->text);
-		wfree(scr->window_menu->entries[MC_SHADE]->text);
-	} else {
-		updateWorkspaceMenu(scr->workspace_submenu);
-	}
-
-	menu = scr->window_menu;
-	if (menu->flags.mapped) {
-		wMenuUnmap(menu);
-		if (menu->entries[0]->clientdata == wwin) {
-			return;
-		}
-	}
-
-	updateMenuForWindow(menu, wwin);
+	menu = open_window_menu_core(wwin);
+	if (!menu)
+		return;
 
 	x -= menu->frame->core->width / 2;
 
@@ -709,6 +843,7 @@ void DestroyWindowMenu(WScreen *scr)
 		scr->window_menu->entries[MC_MINIATURIZE]->text = NULL;
 		scr->window_menu->entries[MC_MAXIMIZE]->text = NULL;
 		scr->window_menu->entries[MC_SHADE]->text = NULL;
+		scr->window_menu->entries[MC_SELECT]->text = NULL;
 		wMenuDestroy(scr->window_menu, True);
 		scr->window_menu = NULL;
 	}
